# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.os_util import gi_import
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.util.env import first_time, envbool
from xpra.client.gtk3.window.stub_window import GtkStubWindow
from xpra.gtk.keymap import KEY_TRANSLATIONS
from xpra.keyboard.common import KeyEvent
from xpra.log import Logger

Gdk = gi_import("Gdk")

log = Logger("window", "keyboard")

UNICODE_KEYNAMES = envbool("XPRA_UNICODE_KEYNAMES", False)


class KeyboardWindow(GtkStubWindow):

    def init_window(self, client, metadata: typedict, client_props: typedict) -> None:
        self.connect("key-press-event", self.handle_key_press_event)
        self.connect("key-release-event", self.handle_key_release_event)

    def get_window_event_mask(self) -> Gdk.EventMask:
        return Gdk.EventMask.KEY_PRESS_MASK | Gdk.EventMask.KEY_RELEASE_MASK

    def next_keyboard_layout(self, update_platform_layout) -> None:
        self._client.next_keyboard_layout(update_platform_layout)

    def keyboard_layout_changed(self, *args) -> None:
        # used by win32 hooks to tell us about keyboard layout changes for this window
        log("keyboard_layout_changed%s", args)
        self._client.window_keyboard_layout_changed(self)

    def parse_key_event(self, event, pressed: bool) -> KeyEvent:
        keyval = event.keyval
        keycode = event.hardware_keycode
        keyname = Gdk.keyval_name(keyval) or ""
        keyname = KEY_TRANSLATIONS.get((keyname, keyval, keycode), keyname)
        if keyname.startswith("U+") and not UNICODE_KEYNAMES:
            # workaround for MS Windows, try harder to find a valid key
            # see ticket #3417
            keymap = Gdk.Keymap.get_default()
            r = keymap.get_entries_for_keycode(event.hardware_keycode)
            if r[0]:
                for kc in r[2]:
                    keyname = Gdk.keyval_name(kc)
                    if not keyname.startswith("U+"):
                        break
        key_event = KeyEvent()
        key_event.modifiers = self._client.mask_to_names(event.state)
        key_event.keyname = keyname
        key_event.keyval = keyval or 0
        key_event.keycode = keycode
        key_event.group = event.group
        key_event.pressed = pressed
        key_event.string = ""
        try:
            codepoint = Gdk.keyval_to_unicode(keyval)
            key_event.string = chr(codepoint)
        except ValueError as e:
            log(f"failed to parse unicode string value of {event}", exc_info=True)
            try:
                key_event.string = event.string or ""
            except UnicodeDecodeError as ve:
                if first_time(f"key-{keycode}-{keyname}"):
                    log("parse_key_event(%s, %s)", event, pressed, exc_info=True)
                    log.warn("Warning: failed to parse string for key")
                    log.warn(f" {keyname=}, {keycode=}")
                    log.warn(f" {keyval=}, group={event.group}")
                    log.warn(" modifiers=%s", csv(key_event.modifiers))
                    log.warn(f" {e}")
                    log.warn(f" {ve}")
        log("parse_key_event(%s, %s)=%s", event, pressed, key_event)
        return key_event

    def handle_key_press_event(self, _window, event) -> bool:
        key_event = self.parse_key_event(event, True)
        self._client.handle_key_action(self, key_event)
        return True

    def handle_key_release_event(self, _window, event) -> bool:
        key_event = self.parse_key_event(event, False)
        self._client.handle_key_action(self, key_event)
        return True



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any
from collections.abc import Sequence

from xpra.client.base.stub import StubClientMixin
from xpra.client.gui.keyboard_helper import KeyboardHelper
from xpra.keyboard.common import KeyEvent, DELAY_KEYBOARD_DATA
from xpra.util.objects import typedict
from xpra.common import noop
from xpra.log import Logger

log = Logger("keyboard")


def noauto(val: str | Sequence | None) -> str | Sequence | None:
    default = [] if isinstance(val, Sequence) else None
    if not val:
        return default
    if str(val).lower() == "auto":
        return default
    return val


class KeyboardClient(StubClientMixin):
    """
    Utility mixin for clients that handle keyboard input
    """
    PREFIX = "keyboard"

    def __init__(self):
        self.keyboard_enabled = True
        self.keyboard_helper_class: type = KeyboardHelper
        self.keyboard_helper = None
        self.keyboard_grabbed: bool = False
        self.keyboard_sync: bool = False
        self.key_repeat_delay = -1
        self.key_repeat_interval = -1
        self.server_keyboard: bool = True
        self.kh_warning: bool = False

    def init_ui(self, opts) -> None:
        send_keyboard = noop
        if not self.readonly:
            def do_send_keyboard(*parts):
                self.after_handshake(self.send, *parts)

            send_keyboard = do_send_keyboard
        try:
            kwargs = dict(
                (x, noauto(getattr(opts, "keyboard_%s" % x))) for x in (
                    "backend", "model", "layout", "layouts", "variant", "variants", "options",
                )
            )
            self.keyboard_helper = self.keyboard_helper_class(send_keyboard, opts.keyboard_sync,
                                                              opts.shortcut_modifiers,
                                                              opts.key_shortcut,
                                                              opts.keyboard_raw, **kwargs)
            if DELAY_KEYBOARD_DATA and not self.readonly:
                self.after_handshake(self.keyboard_helper.send_keymap)
        except ImportError as e:
            log("error instantiating %s", self.keyboard_helper_class, exc_info=True)
            log.warn(f"Warning: no keyboard support, {e}")

    def cleanup(self) -> None:
        kh = self.keyboard_helper
        if kh:
            self.keyboard_helper = None
            kh.cleanup()

    def get_info(self) -> dict[str, dict[str, Any]]:
        return {KeyboardClient.PREFIX: self.get_keyboard_caps()}

    def get_caps(self) -> dict[str, Any]:
        # return {KeyboardClient.PREFIX: caps}
        return self.get_keyboard_caps()

    def parse_server_capabilities(self, c: typedict) -> bool:
        self.server_keyboard = c.boolget("keyboard", True)
        if not self.server_keyboard and self.keyboard_helper:
            # swallow packets:
            self.keyboard_helper.send = noop
        try:
            from xpra import keyboard
            assert keyboard
        except ImportError:
            log.warn("Warning: keyboard module is missing")
            self.keyboard_enabled = False
            return True
        return True

    def process_ui_capabilities(self, caps: typedict) -> None:
        log("process_ui_capabilities()")
        if self.keyboard_helper:
            modifier_keycodes = caps.dictget("modifier_keycodes", {})
            if modifier_keycodes:
                self.keyboard_helper.set_modifier_mappings(modifier_keycodes)
        self.key_repeat_delay, self.key_repeat_interval = caps.intpair("key_repeat", (-1, -1))

    def get_keyboard_caps(self) -> dict[str, Any]:
        caps = {}
        kh = self.keyboard_helper
        if self.readonly or not kh:
            # don't bother sending keyboard info, as it won't be used
            caps["keyboard"] = False
        else:
            caps["keyboard"] = True
            caps["ibus"] = True
            caps["modifiers"] = self.get_current_modifiers()
            skip = ("keycodes", "x11_keycodes") if DELAY_KEYBOARD_DATA else ()
            caps["keymap"] = kh.get_keymap_properties(skip)
            # show the user a summary of what we have detected:
            self.keyboard_helper.log_keyboard_info()
            delay_ms, interval_ms = kh.key_repeat_delay, kh.key_repeat_interval
            if delay_ms > 0 and interval_ms > 0:
                caps["key_repeat"] = (delay_ms, interval_ms)
            else:
                # cannot do keyboard_sync without a key repeat value!
                # (maybe we could just choose one?)
                kh.keyboard_sync = False
            caps["keyboard_sync"] = kh.sync
        log("keyboard capabilities: %s", caps)
        return caps

    def next_keyboard_layout(self, update_platform_layout) -> None:
        if self.keyboard_helper:
            self.keyboard_helper.next_layout(update_platform_layout)

    def window_keyboard_layout_changed(self, window=None) -> None:
        # win32 can change the keyboard mapping per window...
        log("window_keyboard_layout_changed(%s)", window)
        if self.keyboard_helper:
            self.keyboard_helper.keymap_changed()

    def handle_key_action(self, window, key_event: KeyEvent) -> bool:
        kh = self.keyboard_helper
        if not kh:
            return False
        wid = self._window_to_id[window]
        log(f"handle_key_action({window}, {key_event}) wid={wid:#x}")
        if kh.key_handled_as_shortcut(window, key_event.keyname, key_event.modifiers, key_event.pressed):
            return False
        if self.readonly:
            return False
        kh.process_key_event(wid, key_event)
        return False

    def mask_to_names(self, mask) -> list[str]:
        if self.keyboard_helper is None:
            return []
        return self.keyboard_helper.mask_to_names(int(mask))



# This file is part of Xpra.
# Copyright (C) 2023 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.net.common import HttpResponse

EXTENSION_TO_MIMETYPE: dict[str, str] = {
    ".wasm": "application/wasm",
    ".js": "text/javascript",
    ".css": "text/css",
}


def http_response(content, content_type: str = "text/plain") -> HttpResponse:
    if not content:
        return 404, {}, b""
    if isinstance(content, str):
        content = content.encode("latin1")
    return 200, {
        "Content-type": content_type,
        "Content-Length": len(content),
    }, content


def http_status_request(_uri: str, _post_data: bytes) -> HttpResponse:
    return http_response("ready")


def json_response(data) -> HttpResponse:
    import json  # pylint: disable=import-outside-toplevel
    return http_response(json.dumps(data), "application/json")



# This file is part of Xpra.
# Copyright (C) 2018 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import re
import sys
from time import sleep, monotonic
from typing import Any, NoReturn
from collections.abc import Sequence

from xpra.net.ssh.paramiko.util import keymd5, get_key_fingerprints, load_private_key, SSHSocketConnection
from xpra.scripts.main import InitException, InitExit, shellquote
from xpra.net.connect import host_target_string
from xpra.platform.paths import get_ssh_known_hosts_files
from xpra.platform.info import get_username
from xpra.util.parsing import TRUE_OPTIONS, str_to_bool
from xpra.scripts.pinentry import input_pass, confirm
from xpra.net.ssh.util import get_default_keyfiles
from xpra.net.bytestreams import SOCKET_TIMEOUT
from xpra.exit_codes import ExitCode
from xpra.util.io import stderr_print, umask_context
from xpra.common import noerr
from xpra.util.str_fn import csv
from xpra.util.env import envint, envbool, envfloat, first_time
from xpra.log import Logger

# pylint: disable=import-outside-toplevel

log = Logger("network", "ssh")

from paramiko.ssh_exception import SSHException, ProxyCommandFailure  # noqa: E402
from paramiko.transport import Transport  # noqa: E402
from paramiko import SSHConfig  # noqa: E402

WINDOW_SIZE = envint("XPRA_SSH_WINDOW_SIZE", 2 ** 27 - 1)
TIMEOUT = envint("XPRA_SSH_TIMEOUT", 60)

VERIFY_HOSTKEY = envbool("XPRA_SSH_VERIFY_HOSTKEY", True)
VERIFY_STRICT = envbool("XPRA_SSH_VERIFY_STRICT", False)
ADD_KEY = envbool("XPRA_SSH_ADD_KEY", True)
# which authentication mechanisms are enabled with paramiko:
NONE_AUTH = envbool("XPRA_SSH_NONE_AUTH", True)
PASSWORD_AUTH = envbool("XPRA_SSH_PASSWORD_AUTH", True)
AGENT_AUTH = envbool("XPRA_SSH_AGENT_AUTH", True)
KEY_AUTH = envbool("XPRA_SSH_KEY_AUTH", True)
PASSWORD_RETRY = envint("XPRA_SSH_PASSWORD_RETRY", 2)
SSH_AGENT = envbool("XPRA_SSH_AGENT", False)
BANNER = envbool("XPRA_SSH_BANNER", True)
assert PASSWORD_RETRY >= 0
LOG_FAILED_CREDENTIALS = envbool("XPRA_LOG_FAILED_CREDENTIALS", False)
TEST_COMMAND_TIMEOUT = envint("XPRA_SSH_TEST_COMMAND_TIMEOUT", 10)
EXEC_STDOUT_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDOUT_TIMEOUT", 2)
EXEC_STDERR_TIMEOUT = envfloat("XPRA_SSH_EXEC_STDERR_TIMEOUT", 0)

MSYS_DEFAULT_PATH = os.environ.get("XPRA_MSYS_DEFAULT_PATH", "/mingw64/bin/xpra")
CYGWIN_DEFAULT_PATH = os.environ.get("XPRA_CYGWIN_DEFAULT_PATH", "/cygdrive/c/Program Files/Xpra/Xpra_cmd.exe")
DEFAULT_WIN32_INSTALL_PATH = "C:\\Program Files\\Xpra"
WIN_OPENSSH_AGENT = envbool("XPRA_WIN_OPENSSH_AGENT", False)
WIN_OPENSSH_AGENT_MODULE = "win_openssh.OpenSSHAgentConnection"

PARAMIKO_SESSION_LOST = "No existing session"

WIN32_REGISTRY_QUERY = "REG QUERY \"HKEY_LOCAL_MACHINE\\Software\\Xpra\" /v InstallPath"

AUTH_MODES: Sequence[str] = os.environ.get("XPRA_PARAMIKO_AUTH_MODES", "none,agent,key,password").split(",")

if BANNER:
    from paramiko import auth_handler

    class AuthHandlerOverride(auth_handler.AuthHandler):

        def _parse_userauth_banner(self, m):
            banner = m.get_string()
            self.banner = banner
            if banner:
                try:
                    text = banner.decode("utf8")
                except UnicodeDecodeError:
                    text = repr(banner)
                log.info("SSH Authentication Banner:")
                for msg in text.splitlines():
                    log.info(f" {msg!r}")

    auth_handler.AuthHandler = AuthHandlerOverride
    from paramiko import transport
    transport.AuthHandler = AuthHandlerOverride


class SSHProxyCommandConnection(SSHSocketConnection):
    def __init__(self, ssh_channel, peername, target, info):
        super().__init__(ssh_channel, None, "", peername, target, info)
        self.process = None

    def error_is_closed(self, e) -> bool:
        p = self.process
        if p:
            # if the process has terminated,
            # then the connection must be closed:
            if p[0].poll() is not None:
                return True
        return super().error_is_closed(e)

    def get_socket_info(self) -> dict[str, Any]:
        p = self.process
        if not p:
            return {}
        proc, _ssh, cmd = p
        return {
            "process": {
                "pid": proc.pid,
                "returncode": proc.returncode,
                "command": cmd,
            }
        }

    def close(self) -> None:
        try:
            super().close()
        except OSError:
            # this can happen if the proxy command gets a SIGINT,
            # it's closed already and we don't care
            log("SSHProxyCommandConnection.close()", exc_info=True)


def safe_lookup(config_obj, hostname: str) -> dict:
    try:
        _lookup = getattr(config_obj, "_lookup", None)
        import paramiko
        # older versions don't have the same signature for `_lookup`:
        if _lookup and getattr(paramiko, "__version_info__", (0, )) >= (3, 3):
            # completely duplicate the paramiko logic since they're unwilling to merge a trivial change :(
            options = _lookup(hostname=hostname)
            # Inject HostName if it was not set (this used to be done incidentally
            # during tokenization, for some reason).
            if "hostname" not in options:
                options["hostname"] = hostname
            # Handle canonicalization
            canon = options.get("canonicalizehostname", None) in ("yes", "always")
            maxdots = int(options.get("canonicalizemaxdots", 1))
            if canon and hostname.count(".") <= maxdots:
                # NOTE: OpenSSH manpage does not explicitly state this, but its
                # implementation for CanonicalDomains is 'split on any whitespace'.
                domains = options.get("canonicaldomains", "").split()
                hostname = config_obj.canonicalize(hostname, options, domains)
                # Overwrite HostName again here (this is also what OpenSSH does)
                options["hostname"] = hostname
                options = _lookup(
                    hostname, options, canonical=True, final=True
                )
            else:
                options = _lookup(
                    hostname, options, canonical=False, final=True
                )
        else:
            options = config_obj.lookup(hostname)
        return dict(options or {})
    except ImportError as e:
        log("%s.lookup(%s)", config_obj, hostname, exc_info=True)
        log.warn(f"Warning: unable to load SSH host config for {hostname!r}:")
        log.warn(f" {e}")
        if isinstance(e, ModuleNotFoundError):
            log.warn(" (looks like a 'paramiko' distribution packaging issue)")
    except KeyError as e:
        log("%s.lookup(%s)", config_obj, hostname, exc_info=True)
        log.info(f"paramiko ssh config lookup error for host {hostname!r}")
        if first_time("paramiko-#2338"):
            log.info(" %s: %s", type(e), e)
            log.info(" the paramiko project looks unmaintained:")
            log.info(" https://github.com/paramiko/paramiko/pull/2338")
    return {}


def load_ssh_config() -> SSHConfig:
    ssh_config = SSHConfig()
    etc = "etc" if sys.prefix == "/usr" else sys.prefix + "/etc"
    for config_file in (
            f"{etc}/ssh/ssh_config",
            os.path.expanduser("~/.ssh/config"),
    ):
        if not os.path.exists(config_file):
            continue
        with open(config_file, encoding="utf8") as f:
            try:
                ssh_config.parse(f)
            except Exception as e:
                log(f"parse({config_file})", exc_info=True)
                log.error(f"Error parsing {config_file!r}:")
                log.estr(e)
        log(f"parsed ssh config {config_file!r}")
    try:
        log("%i hosts found", len(ssh_config.get_hostnames()))
    except KeyError:
        pass
    return ssh_config


def connect_to(display_desc: dict) -> SSHSocketConnection:
    log(f"connect_to({display_desc})")
    # plain socket attributes:
    host: str = display_desc["host"]
    port: int = display_desc.get("port", 0)
    # ssh and command attributes:
    default_username = get_username()
    username = display_desc.get("username", "")
    if "proxy_host" in display_desc:
        display_desc.setdefault("proxy_username", get_username())
    password: str = display_desc.get("password", "")
    remote_xpra = display_desc["remote_xpra"]
    proxy_command: Sequence[str] = display_desc["proxy_command"]  # ie: ["_proxy_start"]
    socket_dir: str = display_desc.get("socket_dir", "")
    display: str = display_desc.get("display", "")
    display_as_args: list[str] = display_desc["display_as_args"]  # ie: "--start=xterm :10"
    paramiko_config: dict = display_desc.copy()
    paramiko_config.update(display_desc.get("paramiko-config", {}))

    def fail(msg: str) -> NoReturn:
        log("connect_to(%s)", display_desc, exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, msg) from None

    ssh_config = load_ssh_config()

    def ssh_lookup(key) -> dict:
        return safe_lookup(ssh_config, key)

    host_config: dict = ssh_lookup("*")
    host_config.update(ssh_lookup(host))
    log(f"host_config({host})={host_config}")

    # paramiko does not strip comments from configs:
    # https://github.com/Xpra-org/xpra/issues/4503
    def safeget(key: str, default_value: str | int | list) -> Any:
        value = host_config.get(key, default_value)
        if isinstance(value, str):
            value = value.split("#", 1)[0]
        elif isinstance(value, list):
            value = [str(v).split("#", 1)[0] for v in value]
        return value

    def configvalue(key: str, default_value):
        # if the paramiko config has a setting, honour it:
        if paramiko_config and key in paramiko_config:
            return paramiko_config.get(key)
        # fallback to the value from the host config:
        return safeget(key, default_value)

    def configbool(key: str, default_value=True) -> bool:
        return str_to_bool(configvalue(key, default_value), default_value)

    def get_keyfiles(config_name="key") -> list[str]:
        keyfiles: list[str] = safeget("identityfile", [])
        keyfile = paramiko_config.get(config_name, "")
        if keyfile:
            keyfiles.insert(0, keyfile)
        if not configbool("identitiesonly", False):
            keyfiles += get_default_keyfiles()
        return keyfiles

    if host_config:
        log(f"got host config for {host!r}: {host_config}")
        host = safeget("hostname", host)
        # don't override the username specified on the display string
        username = username or safeget("user", username) or default_username
        port = port or safeget("port", port) or 22
        paramiko_config["port"] = port  # ensure paramiko_config has the same port
        try:
            port = int(port)
        except (TypeError, ValueError):
            raise InitExit(ExitCode.SSH_FAILURE, f"invalid ssh port specified: {port!r}") from None
        proxycommand = host_config.get("proxycommand", "")
        if proxycommand:
            log(f"found proxycommand={proxycommand!r} for host {host!r}")
            from paramiko import ProxyCommand
            sock = ProxyCommand(proxycommand)
            log(f"ProxyCommand({proxycommand})={sock}")
            from xpra.util.child_reaper import get_child_reaper
            cmd = getattr(sock, "cmd", [])

            def proxycommand_ended(proc) -> None:
                log(f"proxycommand_ended({proc}) exit code={proc.poll()}")

            get_child_reaper().add_process(sock.process, "paramiko-ssh-client", cmd, True, True,
                                           callback=proxycommand_ended)
            proxy_keys = get_keyfiles("proxy_key")
            log(f"{proxy_keys=}")
            from paramiko.client import SSHClient
            ssh_client = SSHClient()
            ssh_client.load_system_host_keys()
            log("ssh proxy command connect to %s", (host, port, sock))
            ssh_client.connect(host, port, sock=sock)
            transport = ssh_client.get_transport()
            do_connect_to(transport, host, port,
                          username, password,
                          host_config,
                          proxy_keys,
                          paramiko_config)
            chan = run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args, paramiko_config)
            peername = (host, port)
            conn = SSHProxyCommandConnection(chan, peername, peername, {"host": host, "port": port})
            conn.target = host_target_string("ssh", username, host, port or 22, display)
            conn.timeout = SOCKET_TIMEOUT
            conn.start_stderr_reader()
            conn.process = (sock.process, "ssh", cmd)
            from xpra.net import bytestreams
            bytestreams.CLOSED_EXCEPTIONS = tuple(list(bytestreams.CLOSED_EXCEPTIONS) + [ProxyCommandFailure])
            return conn

    keys = get_keyfiles()
    from xpra.net.socket_util import socket_connect
    if "proxy_host" in display_desc:
        proxy_host = display_desc["proxy_host"]
        proxy_port = int(display_desc.get("proxy_port", 22))
        proxy_username = display_desc.get("proxy_username", username) or default_username
        proxy_password = display_desc.get("proxy_password", password)
        proxy_keys = get_keyfiles("proxy_key")
        sock = socket_connect(proxy_host, proxy_port)
        if not sock:
            fail(f"SSH proxy transport failed to connect to {proxy_host}:{proxy_port}")
        middle_transport = do_connect(sock, proxy_host, proxy_port,
                                      proxy_username, proxy_password,
                                      ssh_lookup(host) or ssh_lookup("*"),
                                      proxy_keys,
                                      paramiko_config)
        log("Opening proxy channel")
        chan_to_middle = middle_transport.open_channel("direct-tcpip", (host, port), ("localhost", 0))
        transport = do_connect(chan_to_middle, host, port,
                               username or default_username, password,
                               host_config,
                               keys,
                               paramiko_config)
        chan = run_remote_xpra(transport, proxy_command, remote_xpra, socket_dir, display_as_args, paramiko_config)
        peername = (proxy_host, proxy_port)
        conn = SSHProxyCommandConnection(chan, peername, peername, {"host": proxy_host, "port": proxy_port})
        to_str = host_target_string("ssh", username or default_username, host, port, display)
        proxy_str = host_target_string("ssh", proxy_username, proxy_host, proxy_port)
        conn.target = f"{to_str} via {proxy_str}"
        conn.timeout = SOCKET_TIMEOUT
        conn.start_stderr_reader()
        return conn

    # plain TCP connection to the server,
    # we open it then give the socket to paramiko:
    auth_modes = get_auth_modes(paramiko_config, host_config, password)
    log(f"authentication modes={auth_modes}")
    sock = None
    transport = None
    sockname = host
    username = username or default_username
    peername = ()
    port = port or 22
    while not transport:
        log("opening socket connection to %r:%i", host, port)
        sock = socket_connect(host, port)
        if not sock:
            fail(f"SSH failed to connect to {host}:{port}")
        sockname = sock.getsockname()
        peername = sock.getpeername()
        log(f"paramiko socket_connect: sockname={sockname}, peername={peername}")
        try:
            transport = do_connect(sock, host, port,
                                   username, password,
                                   host_config,
                                   keys,
                                   paramiko_config,
                                   auth_modes)
        except SSHAuthenticationError as e:
            log(f"paramiko authentication errors on socket {sock} with modes {auth_modes}: {e.errors}",
                exc_info=True)
            pw_errors = []
            for errs in e.errors.values():
                pw_errors += errs
            if ("key" in auth_modes or "agent" in auth_modes) and PARAMIKO_SESSION_LOST in pw_errors:
                # try connecting again but without 'key' and 'agent' authentication:
                # see https://github.com/Xpra-org/xpra/issues/3223
                for m in ("key", "agent"):
                    try:
                        auth_modes.remove(m)
                    except KeyError:
                        pass
                log.info(f"retrying SSH authentication with modes {csv(auth_modes)}")
                continue
            raise
        finally:
            if sock and not transport:
                noerr(sock.shutdown)
                noerr(sock.close)
                sock = None

    remote_port: int = display_desc.get("remote_port", 0)
    if remote_port:
        # we want to connect directly to a remote port,
        # we don't need to run a command
        chan = transport.open_channel("direct-tcpip", ("localhost", remote_port), ('localhost', 0))
        log(f"direct channel to remote port {remote_port} : {chan}")
    else:
        chan = run_remote_xpra(transport, proxy_command, remote_xpra,
                               socket_dir, display_as_args, paramiko_config)
    conn = SSHSocketConnection(chan, sock, sockname, peername, (host, port), {"host": host, "port": port})
    conn.target = host_target_string("ssh", username, host, port, display)
    conn.timeout = SOCKET_TIMEOUT
    conn.start_stderr_reader()
    log(f"paramiko.client.connect_to({display_desc})={conn}")
    return conn


def get_auth_modes(paramiko_config, host_config: dict, password: str) -> list[str]:
    def configvalue(key: str):
        # if the paramiko config has a setting, honour it:
        if paramiko_config and key in paramiko_config:
            return paramiko_config.get(key)
        # fallback to the value from the host config:
        return host_config.get(key)

    def configbool(key: str, default_value=True) -> bool:
        return str_to_bool(configvalue(key), default_value)

    auth_str = configvalue("auth")
    if auth_str:
        return auth_str.split("+")
    auth = []
    identitiesonly = configbool("identitiesonly", False)
    if not identitiesonly:
        if configbool("noneauthentication", NONE_AUTH):
            auth.append("none")
        if password and configbool("passwordauthentication", PASSWORD_AUTH):
            auth.append("password")
    if configbool("agentauthentication", AGENT_AUTH):
        auth.append("agent")
    # Some people do two-factor using KEY_AUTH to kick things off, so this happens first
    if configbool("keyauthentication", KEY_AUTH):
        auth.append("key")
    if not identitiesonly and not password and configbool("passwordauthentication", PASSWORD_AUTH):
        auth.append("password")
    return auth


class IAuthHandler:
    def __init__(self, password):
        self.authcount = 0
        self.password = password

    def handle_request(self, title: str, instructions, prompt_list) -> list:
        log("handle_request%s counter=%i", (title, instructions, prompt_list), self.authcount)
        p = []
        for pent in prompt_list:
            if self.password:
                p.append(self.password)
                self.password = None
            else:
                p.append(input_pass(pent[0]))
        self.authcount += 1
        log(f"handle_request(..) returning {len(p)} values")
        return p


def do_connect(chan, host: str, port: int,
               username: str, password: str,
               host_config: dict, keyfiles: list[str], paramiko_config: dict, auth_modes=AUTH_MODES):
    transport = Transport(chan)
    transport.use_compression(False)
    log("SSH transport %s", transport)
    try:
        transport.start_client()
    except SSHException as e:
        log("SSH negotiation failed", exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, "SSH negotiation failed: %s" % e) from None
    return do_connect_to(transport, host, port,
                         username, password,
                         host_config, keyfiles, paramiko_config, auth_modes)


def do_connect_to(transport, host: str, port: int,
                  username: str, password: str,
                  host_config, keyfiles: Sequence[str], paramiko_config: dict, auth_modes=AUTH_MODES):
    AuthenticationManager(transport, host, port,
                          username, password, host_config, keyfiles, paramiko_config, auth_modes).run()
    return transport


def load_host_keys() -> tuple[str, Any]:   # returns a HostKeys object
    from paramiko.hostkeys import HostKeys
    host_keys = HostKeys()
    host_keys_filename = ""
    known_hosts_files = get_ssh_known_hosts_files()
    for known_hosts in known_hosts_files:
        host_keys.clear()
        try:
            path = os.path.expanduser(known_hosts)
            if os.path.exists(path):
                host_keys.load(path)
                log("HostKeys.load(%s) successful", path)
                host_keys_filename = path
                break
        except OSError:
            log("HostKeys.load(%s)", known_hosts, exc_info=True)
    log("host keys=%s", host_keys)
    return host_keys_filename, host_keys


def verify_hostkey(host: str, host_key, verifyhostkeydns: bool, stricthostkeychecking: bool, addkey: bool) -> str:
    host_keys_filename, host_keys = load_host_keys()
    keys = safe_lookup(host_keys, host)
    known_host_key = keys.get(host_key.get_name())

    def keyname() -> str:
        return host_key.get_name().replace("ssh-", "")

    if known_host_key and host_key == known_host_key:
        assert host_key
        log("%s host key '%s' OK for host '%s'", keyname(), keymd5(host_key), host)
        return ""

    dnscheck = ""
    if verifyhostkeydns:
        try:
            from xpra.net.ssh.sshfp import check_host_key
            dnscheck = check_host_key(host, host_key)
        except ImportError as e:
            log("verifyhostkeydns failed", exc_info=True)
            log.info("cannot check SSHFP DNS records")
            log.info(" %s", e)
    log("dnscheck=%s", dnscheck)

    def adddnscheckinfo(q: list[str]):
        if dnscheck is not True:
            if dnscheck:
                q.append("SSHFP validation failed:")
                q.append(str(dnscheck))
            else:
                q.append("SSHFP validation failed")

    if dnscheck is True:
        # DNSSEC provided a matching record
        log.info("found a valid SSHFP record for host %s", host)
    elif known_host_key:
        log.warn("Warning: SSH server key mismatch")
        qinfo: list[str] = [
            "WARNING: REMOTE HOST IDENTIFICATION HAS CHANGED!",
            "IT IS POSSIBLE THAT SOMEONE IS DOING SOMETHING NASTY!",
            "Someone could be eavesdropping on you right now (man-in-the-middle attack)!",
            "It is also possible that a host key has just been changed.",
            f"The fingerprint for the {keyname()} key sent by the remote host is",
            keymd5(host_key),
        ]
        adddnscheckinfo(qinfo)
        if stricthostkeychecking:
            log.warn("Host key verification failed.")
            # TODO: show alert with no option to accept key
            qinfo += [
                "Please contact your system administrator.",
                "Add correct host key to get rid of this message.",
                f"Offending {keyname()} key in {host_keys_filename}",
                f"ECDSA host key for {keyname()} has changed and you have requested strict checking.",
            ]
            stderr_print(os.linesep.join(qinfo))
            return "SSH Host key has changed"
        if not confirm(qinfo):
            return "SSH Host key has changed"
        log.info("host key confirmed")
    else:
        assert (not keys) or (host_key.get_name() not in keys)
        if not keys:
            log.warn("Warning: unknown SSH host '%s'", host)
        else:
            log.warn("Warning: unknown %s SSH host key", keyname())
        qinfo = [
            f"The authenticity of host {host!r} can't be established.",
            f"{keyname()} key fingerprint is",
            keymd5(host_key),
        ]
        adddnscheckinfo(qinfo)
        if not confirm(qinfo):
            return f"Unknown SSH host {host!r}"
        log.info("host key confirmed")
    if not addkey:
        # we're done!
        return ""
    known_hosts_files = get_ssh_known_hosts_files()
    if not known_hosts_files:
        # we can't save it
        return ""
    if addkey:
        host_keys.add(host, host_key.get_name(), host_key)
        save_host_key(host_keys, host_keys_filename)
    return ""


def save_host_key(host_keys, host_keys_filename: str) -> None:
    filenames = [host_keys_filename]
    for filename in get_ssh_known_hosts_files():
        if filename not in filenames:
            filenames.append(filename)
    for filename in filenames:
        try:
            log(f"adding key to {filename!r}")
            if not os.path.exists(filename):
                keys_dir = os.path.dirname(filename)
                if not os.path.exists(keys_dir):
                    log(f"creating keys directory {keys_dir!r}")
                    os.mkdir(keys_dir, 0o700)
                elif not os.path.isdir(keys_dir):
                    log.warn(f"Warning: {keys_dir!r} is not a directory")
                    log.warn(f" key not saved to {filename!r}")
                    continue
                if os.path.exists(keys_dir) and os.path.isdir(keys_dir):
                    log(f"creating known host file {host_keys_filename!r}")
                    with umask_context(0o133):
                        with open(host_keys_filename, "ab+"):
                            "file has been created"
            host_keys.save(host_keys_filename)
            return
        except OSError as e:
            log(f"failed to add key to {host_keys_filename!r}")
            log.error(f"Error adding key to {host_keys_filename!r}")
            log.error(f" {e}")


class AuthenticationManager:
    def __init__(self, transport, host: str, port: int, username: str, password: str,
                 host_config, keyfiles: Sequence[str], paramiko_config: dict, auth_modes=AUTH_MODES):
        log("AuthenticationManager%s", (transport, host, username, "..", host_config, keyfiles, paramiko_config))
        self.transport = transport
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.host_config = host_config
        self.keyfiles = keyfiles
        self.paramiko_config = paramiko_config
        self.auth_modes = auth_modes
        self.auth_errors: dict[str, list[str]] = {}

    def configvalue(self, key: str) -> Any:
        # if the paramiko config has a setting, honour it:
        if key in self.paramiko_config:
            return self.paramiko_config.get(key)
        # fallback to the value from the host config:
        return (self.host_config or {}).get(key)

    def configbool(self, key: str, default_value=True) -> bool:
        return str_to_bool(self.configvalue(key), default_value)

    def configint(self, key: str, default_value=0) -> int:
        v = self.configvalue(key)
        if v is None:
            return default_value
        return int(v)

    def verify_hostkey(self, host_key):
        verifyhostkeydns = self.configbool("verifyhostkeydns", True)
        stricthostkeychecking = self.configbool("stricthostkeychecking", VERIFY_STRICT)
        addkey = self.configbool("addkey", ADD_KEY)
        port = self.configint("port", self.port)
        host = self.host
        if port != 22:
            host = f"[{host}]:{port}"  # this is how the host is stored in known_hosts for non-standard ports
        err = verify_hostkey(host, host_key, verifyhostkeydns, stricthostkeychecking, addkey)
        if err:
            self.transport.close()
            raise InitExit(ExitCode.SSH_KEY_FAILURE, err)

    def run(self) -> None:
        host_key = self.transport.get_remote_server_key()
        assert host_key, "no remote server key"
        log("remote_server_key=%s", keymd5(host_key))
        if self.configbool("verify-hostkey", VERIFY_HOSTKEY):
            self.verify_hostkey(host_key)
        else:
            log("ssh host key verification skipped")

        banner = self.transport.get_banner()
        if banner:
            log.info("SSH server banner:")
            for x in banner.splitlines():
                log.info(f" {x}")

        log(f"starting authentication, authentication methods: {self.auth_modes}")
        auth = list(self.auth_modes)
        # per the RFC we probably should do none first always and read off the supported
        # methods, however, the current code seems to work fine with OpenSSH
        while not self.transport.is_authenticated() and auth:
            a = auth.pop(0)
            log("auth=%s", a)
            if a == "none":
                self.auth_none()
            elif a == "agent":
                self.auth_agent()
            elif a == "key":
                self.auth_publickey()
            elif a == "password":
                self.auth_password()
            else:
                log.warn(f"Warning: invalid authentication mechanism {a}")
            # detect session-lost problems:
            # (no point in continuing without a session)
            if self.auth_errors and not self.transport.is_authenticated():
                for err_strs in self.auth_errors.values():
                    if PARAMIKO_SESSION_LOST in err_strs:
                        raise SSHAuthenticationError(self.host, self.auth_errors)
        if not self.transport.is_authenticated():
            self.transport.close()
            log(f"authentication errors: {self.auth_errors}")
            raise SSHAuthenticationError(self.host, self.auth_errors)

    def auth_none(self) -> None:
        log("trying none authentication")
        try:
            self.transport.auth_none(self.username)
        except SSHException as e:
            self.auth_errors.setdefault("none", []).append(str(e))
            log("auth_none()", exc_info=True)

    def auth_agent(self) -> None:
        from paramiko.agent import Agent
        allowed_key_fingerprints = get_key_fingerprints(self.keyfiles)
        agent = Agent()
        all_agent_keys = agent.get_keys()
        log("agent keys: %s", all_agent_keys)
        log("allowed key fingerprints: %s", allowed_key_fingerprints)
        agent_keys = [x for x in all_agent_keys if x.get_fingerprint() in allowed_key_fingerprints]
        log("agent keys matching fingerprints: %s", agent_keys)
        if not self.configbool("identitiesonly", False):
            for agent_key in all_agent_keys:
                try:
                    if agent_key not in agent_keys:
                        agent_keys.append(agent_key)
                except NotImplementedError:
                    log("auth_agent()", exc_info=True)
                    log.warn("Warning: failed to compare agent keys")
                    agent_keys.append(agent_key)
        if not agent_keys:
            log.info("no ssh agent keys")
            return
        tried = 0
        for agent_key in agent_keys:
            tried += 1
            log("trying ssh-agent key '%s'", keymd5(agent_key))
            try:
                self.transport.auth_publickey(self.username, agent_key)
                log("tried ssh-agent key '%s'", keymd5(agent_key))
                if self.transport.is_authenticated():
                    log("authenticated using agent and key '%s'", keymd5(agent_key))
                    return
            except AttributeError as e:
                log(f"auth_publickey({self.username}, {agent_key})")
                log.warn("Warning: paramiko bug during public key agent authentication")
                log.warn(f" {type(e)}: {e}")
                log.warn(f" using key {type(agent_key)}: {agent_key}")
            except SSHException as e:
                self.auth_errors.setdefault("agent", []).append(str(e))
                log.info("SSH agent key '%s' rejected for user '%s'", keymd5(agent_key), self.username)
                log("%s%s", self.transport.auth_publickey, (self.username, agent_key), exc_info=True)
                if str(e) == PARAMIKO_SESSION_LOST:
                    # no point in trying more keys
                    break
        if not self.transport.is_authenticated():
            log.info("agent authentication failed, tried %i keys", tried)

    def auth_publickey(self) -> None:
        log(f"trying public key authentication using {self.keyfiles}")
        for keyfile_path in self.keyfiles:
            key = load_private_key(keyfile_path)
            if not key:
                if os.path.exists(keyfile_path):
                    log.error(f"Error: cannot load private key {keyfile_path!r}")
                continue
            log(f"auth_publickey using {keyfile_path!r}: {keymd5(key)}")
            try:
                self.transport.auth_publickey(self.username, key)
                if self.transport.is_authenticated():
                    return
            except SSHException as e:
                self.auth_errors.setdefault("key", []).append(str(e))
                log(f"key {keyfile_path!r} rejected", exc_info=True)
                log.info(f"SSH authentication using key {keyfile_path!r} failed:")
                log.info(f" {e}")
                if str(e) == PARAMIKO_SESSION_LOST:
                    # no point in trying more keys
                    return

    def auth_password(self) -> None:
        self.auth_interactive()
        if not self.transport.is_authenticated():
            if self.password:
                self.do_auth_password(self.password)
            else:
                tries = self.configint("numberofpasswordprompts", PASSWORD_RETRY)
                for _ in range(tries):
                    password = input_pass(f"please enter the SSH password for {self.username}@{self.host}")
                    if not password:
                        break
                    self.do_auth_password(password)
                    if self.transport.is_authenticated():
                        break

    def do_auth_password(self, password: str) -> None:
        log("trying password authentication")
        try:
            self.transport.auth_password(self.username, password)
        except SSHException as authe:
            estr = getattr(authe, "message", str(authe))
            self.auth_errors.setdefault("password", []).append(estr)
            log("do_auth_password(..)", exc_info=True)
            emsgs = estr.split(";")
        else:
            emsgs = []
        if not self.transport.is_authenticated():
            log.info("SSH password authentication failed:")
            for emsg in emsgs:
                log.info(f" {emsg}")
            if log.is_debug_enabled() and LOG_FAILED_CREDENTIALS:
                log.info(f" invalid username {self.username!r} or password {password!r}")

    def auth_interactive(self) -> None:
        log("trying interactive authentication")
        try:
            iauthhandler = IAuthHandler(self.password)
            self.transport.auth_interactive(self.username, iauthhandler.handle_request, "")
        except SSHException as authe:
            estr = getattr(authe, "message", str(authe))
            self.auth_errors.setdefault("interactive", []).append(estr)
            log("auth_interactive(..)", exc_info=True)
            log.info("SSH password authentication failed:")
            for emsg in getattr(authe, "message", estr).split(";"):
                log.info(f" {emsg}")


class SSHAuthenticationError(InitExit):
    def __init__(self, host, errors):
        super().__init__(ExitCode.CONNECTION_FAILED, f"SSH Authentication failed for {host!r}")
        self.errors = errors


def run_test_command(transport, cmd: str) -> tuple[list[str], list[str], int]:
    log(f"run_test_command(transport, {cmd})")
    try:
        chan = transport.open_session(window_size=None, max_packet_size=0, timeout=60)
        chan.set_name(f"run-test:{cmd}")
    except SSHException as e:
        log("open_session", exc_info=True)
        raise InitExit(ExitCode.SSH_FAILURE, f"failed to open SSH session: {e}") from None
    chan.exec_command(cmd)
    log(f"exec_command({cmd!r}) returned")
    start = monotonic()
    while not chan.exit_status_ready():
        if monotonic() - start > TEST_COMMAND_TIMEOUT:
            chan.close()
            raise InitException(f"SSH test command {cmd!r} timed out")
        log("exit status is not ready yet, sleeping")
        sleep(0.01)
    code = chan.recv_exit_status()
    log(f"exec_command({cmd!r})={code}")

    def chan_read(fileobj) -> list[str]:
        try:
            return fileobj.readlines()
        except OSError:
            log(f"chan_read({fileobj})", exc_info=True)
            return []
        finally:
            noerr(fileobj.close)

    # don't wait too long for the data:
    chan.settimeout(EXEC_STDOUT_TIMEOUT)
    out = chan_read(chan.makefile())
    log(f"exec_command out={out!r}")
    chan.settimeout(EXEC_STDERR_TIMEOUT)
    err = chan_read(chan.makefile_stderr())
    log(f"exec_command err={err!r}")
    try:
        chan.close()
    except EOFError:
        log("chan.close() raised EOFError - transport already closed")
    return out, err, code


def run_remote_xpra(transport, xpra_proxy_command: Sequence[str], remote_xpra: Sequence[str],
                    socket_dir: str, display_as_args: Sequence[str], paramiko_config: dict):
    log("run_remote_xpra%s", (transport, xpra_proxy_command, remote_xpra,
                              socket_dir, display_as_args, paramiko_config))
    crf = ChannelRunFactory(transport, xpra_proxy_command, remote_xpra, socket_dir, display_as_args, paramiko_config)
    return crf.open()


class ChannelRunFactory:

    def __init__(self, transport, xpra_proxy_command: Sequence[str], remote_xpra: Sequence[str],
                 socket_dir: str, display_as_args: Sequence[str], paramiko_config: dict):
        self.transport = transport
        self.xpra_proxy_command = xpra_proxy_command
        self.remote_xpra = remote_xpra
        self.socket_dir = socket_dir
        self.display_as_args = display_as_args
        self.paramiko_config = paramiko_config

    def rtc(self, cmd: str) -> tuple[list[str], list[str], int]:
        return run_test_command(self.transport, cmd)

    def detectosname(self) -> str:
        # first, try a syntax that should work with any ssh server:
        r = self.rtc("echo %OS%")
        if r[2] == 0 and r[0]:
            name = r[0][-1].rstrip("\n\r")
            log(f"echo %OS%={name!r}")
            if name != "%OS%":
                # MS Windows OS will return "Windows_NT" here
                log.info(f"ssh server OS is {name!r}")
                return name
        # this should work on all other OSes:
        r = self.rtc("echo $OSTYPE")
        if r[2] == 0 and r[0]:
            name = r[0][-1].rstrip("\n\r")
            log(f"OSTYPE={name!r}")
            log.info(f"ssh server OS is {name!r}")
            return name
        return "unknown"

    def getexeinstallpath(self, osname: str) -> str:
        cmd = WIN32_REGISTRY_QUERY
        if osname == "msys":
            # escape for msys shell:
            cmd = cmd.replace("/", "//")
        r = self.rtc(cmd)
        if r[2] != 0:
            return ""
        lines = r[0]
        for line in lines:
            qmatch = re.search(r"InstallPath\s*\w*\s*(.*)", line)
            if qmatch:
                return qmatch.group(1).rstrip("\n\r")
        return ""

    def get_which(self, find_command="which", command="xpra") -> str:
        r = self.rtc(f"{find_command} {command}")
        if r[2] != 0:
            return ""
        lines = r[0]
        if not lines:
            return ""
        # can be multiple lines when the command output is polluted
        for line in lines:
            line = line.rstrip("\n\r ").lstrip("\t ")
            if not line or line == "OK":
                continue
            # use the actual path returned by 'command -v' or 'which':
            if line.startswith(f"alias {command}="):
                # ie: "alias xpra='xpra -d proxy'" -> "xpra -d proxy"
                return line.split("=", 1)[1].strip("'")
            return line
        return ""

    def find_remote_xpra(self) -> str:
        log(f"find_remote_xpra() {self.remote_xpra=}")
        osname = self.detectosname()

        def winpath(p) -> str:
            if osname == "msys":
                return p.replace("\\", "\\\\")
            if osname == "cygwin":
                return "/cygdrive/" + p.replace(":\\", "/").replace("\\", "/")
            return p

        find_command = ""
        for xpra_cmd in self.remote_xpra:
            if osname.startswith("Windows") or osname in ("msys", "cygwin"):
                # on MS Windows,
                # always prefer the application path found in the registry:
                installpath = self.getexeinstallpath(osname)
                if installpath:
                    return winpath(f"{installpath}\\Xpra_cmd.exe")
                elif xpra_cmd.find("/") < 0 and xpra_cmd.find("\\") < 0:
                    test_path = winpath(f"{DEFAULT_WIN32_INSTALL_PATH}\\{xpra_cmd}")
                    cmd = f'dir "{test_path}"'
                    r = self.rtc(cmd)
                    if r[2] == 0:
                        return test_path

            if not find_command and not osname.startswith("Windows"):
                if self.rtc("command")[2] == 0:
                    find_command = "command -v"
                else:
                    find_command = "which"

            if find_command:
                found = self.get_which(find_command, xpra_cmd)
                if found:
                    return found

            if xpra_cmd == "xpra" and osname in ("msys", "cygwin"):
                default_path = CYGWIN_DEFAULT_PATH if osname == "cygwin" else MSYS_DEFAULT_PATH
                if default_path:
                    # try the default system installation path
                    r = self.rtc(f"command -v '{default_path}'")
                    if r[2] == 0:
                        return default_path  # ie: "/mingw64/bin/xpra"
        return ""

    def get_xpra_command(self) -> str:
        log(f"will try to run xpra from: {self.remote_xpra}")
        xpra_cmd = self.find_remote_xpra() or "xpra"
        log(f"adding {xpra_cmd=!r}")
        cmd = '"' + xpra_cmd + '" ' + ' '.join(shellquote(x) for x in self.xpra_proxy_command)
        if self.socket_dir:
            cmd += f" \"--socket-dir={self.socket_dir}\""
        if self.display_as_args:
            cmd += " "
            cmd += " ".join(shellquote(x) for x in self.display_as_args)
        log(f"cmd({self.xpra_proxy_command}, {self.socket_dir}, {self.display_as_args})={cmd}")
        return cmd

    def open(self):
        cmd = self.get_xpra_command()

        # see https://github.com/paramiko/paramiko/issues/175
        # WINDOW_SIZE = 2097152
        log(f"trying to open SSH session, window-size={WINDOW_SIZE}, timeout={TIMEOUT}")
        try:
            chan = self.transport.open_session(window_size=WINDOW_SIZE, max_packet_size=0, timeout=TIMEOUT)
            chan.set_name("run-xpra")
        except SSHException as e:
            log("open_session", exc_info=True)
            raise InitExit(ExitCode.SSH_FAILURE, f"failed to open SSH session: {e}") from None
        agent_option = str(self.paramiko_config.get("agent", SSH_AGENT)) or "no"
        log(f"paramiko {agent_option=}")
        if agent_option.lower() in TRUE_OPTIONS:
            if not WIN_OPENSSH_AGENT and WIN_OPENSSH_AGENT_MODULE not in sys.modules:
                log(f"preventing {WIN_OPENSSH_AGENT_MODULE!r} from loading")
                # noinspection PyTypeChecker
                sys.modules[WIN_OPENSSH_AGENT_MODULE] = None
            log.info("paramiko SSH agent forwarding enabled")
            from paramiko.agent import AgentRequestHandler
            AgentRequestHandler(chan)
        log(f"channel exec_command({cmd!r})")
        chan.exec_command(cmd)
        log("exec_command sent, returning channel for service")
        return chan



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any
from subprocess import Popen

from xpra.os_util import gi_import
from xpra.util.system import is_X11
from xpra.util.version import dict_version_trim
from xpra.util.screen import prettify_plug_name
from xpra.common import FULL_INFO, noop
from xpra.gtk.versions import get_gtk_version_info
from xpra.gtk.info import get_screen_sizes
from xpra.server import features
from xpra.server.subsystem.stub import StubServerMixin
from xpra.log import Logger

log = Logger("server", "gtk")

get_default_window_icon_fallback = noop


def gdk_init() -> None:
    log("gdk_init()")
    try:
        from xpra.x11.gtk.display_source import init_gdk_display_source
    except ImportError as e:
        log.warn(f"Warning: unable to initialize gdk display source: {e}")
        return
    init_gdk_display_source()
    global get_default_window_icon_fallback
    if get_default_window_icon_fallback == noop:
        # inject Gtk into the windowicon lookup:
        try:
            from xpra.server.window import windowicon
            get_default_window_icon_fallback = windowicon.do_get_default_window_icon
            windowicon.get_default_window_icon = get_default_window_icon
        except ImportError:
            pass


def get_default_window_icon(size: int, wmclass_name: str):
    Gtk = gi_import("Gtk")
    it = Gtk.IconTheme.get_default()  # pylint: disable=no-member
    log("get_default_window_icon(%i) icon theme=%s, wmclass_name=%s", size, it, wmclass_name)
    for icon_name in (
            f"{wmclass_name}-color",
            wmclass_name,
            f"{wmclass_name}_{size}x{size}",
            f"application-x-{wmclass_name}",
            f"{wmclass_name}-symbolic",
            f"{wmclass_name}.symbolic",
    ):
        i = it.lookup_icon(icon_name, size, 0)
        log("lookup_icon(%s)=%s", icon_name, i)
        if not i:
            continue
        try:
            pixbuf = i.load_icon()
            log("load_icon()=%s", pixbuf)
            if pixbuf:
                w, h = pixbuf.props.width, pixbuf.props.height
                log("using '%s' pixbuf %ix%i", icon_name, w, h)
                return w, h, "RGBA", pixbuf.get_pixels()
        except Exception:
            log("%s.load_icon()", i, exc_info=True)
    return get_default_window_icon_fallback(size, wmclass_name)


class GTKServer(StubServerMixin):

    def __init__(self):
        StubServerMixin.__init__(self)
        self.xvfb: Popen | None = None
        self.display = os.environ.get("DISPLAY", "")
        self.x11_filter = False

    def init(self, opts) -> None:
        log("GTKServer.init(..)")
        from xpra.scripts.main import no_gtk
        no_gtk()

    def get_display_name(self) -> str:
        Gdk = gi_import("Gdk")
        manager = Gdk.DisplayManager.get()
        display = manager.get_default_display()
        return prettify_plug_name(display.get_name())

    def watch_keymap_changes(self) -> None:
        # Set up keymap change notification:
        from xpra.gtk.keymap import get_default_keymap
        keymap = get_default_keymap()
        keymap.connect("keys-changed", self.keymap_changed)

    def setup(self) -> None:
        if is_X11():
            from xpra.scripts.server import verify_display
            if not verify_display(xvfb=self.xvfb, display_name=self.display):
                from xpra.scripts.config import InitExit
                from xpra.exit_codes import ExitCode
                raise InitExit(ExitCode.NO_DISPLAY, f"unable to access display {self.display!r}")
            gdk_init()
            from xpra.x11.gtk.bindings import init_x11_filter
            self.x11_filter = init_x11_filter()
            assert self.x11_filter
        if features.display:
            Gdk = gi_import("Gdk")
            screen = Gdk.Screen.get_default()
            if screen:
                screen.connect("size-changed", self._screen_size_changed)
                screen.connect("monitors-changed", self._monitors_changed)

    def _screen_size_changed(self, screen) -> None:
        log(f"_screen_size_changed({screen})")
        self.schedule_screen_changed(screen)

    def _monitors_changed(self, screen) -> None:
        log(f"_monitors_changed({screen})")
        self.schedule_screen_changed(screen)

    def cleanup(self) -> None:
        if not self.x11_filter:
            return
        self.x11_filter = False
        from xpra.x11.gtk.bindings import cleanup_x11_filter
        cleanup_x11_filter()

    def late_cleanup(self, stop=True) -> None:
        from xpra.gtk.util import close_gtk_display
        close_gtk_display()
        if is_X11():
            from xpra.x11.gtk.display_source import close_gdk_display_source
            close_gdk_display_source()

    def get_caps(self, source) -> dict[str, Any]:
        caps: dict[str, Any] = {}
        if "versions" in source.wants and FULL_INFO >= 2:
            caps["versions"] = get_gtk_version_info()
        if "features" in source.wants:
            caps["screen_sizes"] = get_screen_sizes()
        return caps

    def get_ui_info(self, _proto, **kwargs) -> dict[str, Any]:
        return {
            "versions": dict_version_trim(get_gtk_version_info()),
            "screen_sizes": get_screen_sizes(),
        }



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any

from xpra.util.objects import typedict
from xpra.util.str_fn import bytestostr
from xpra.util.env import OSEnvContext
from xpra.util.version import parse_version, dict_version_trim
from xpra.util.parsing import TRUE_OPTIONS, FALSE_OPTIONS
from xpra.common import FULL_INFO
from xpra.server.subsystem.stub import StubServerMixin
from xpra.log import Logger

log = Logger("opengl")


def run_opengl_probe(cmd: list[str], env: dict[str, str], display_name: str):
    props: dict[str, Any] = {}
    try:
        # pylint: disable=import-outside-toplevel
        from subprocess import Popen, PIPE
        # we want the output so we can parse it:
        env["XPRA_REDIRECT_OUTPUT"] = "0"
        env["XPRA_OPENGL_ZEROCOPY_UPLOAD_WARNING"] = "0"
        log(f"query_opengl() using {cmd=}, {env=}")
        proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, env=env)
        out, err = proc.communicate()
        log("out(%s)=%s", cmd, out)
        log("err(%s)=%s", cmd, err)
        if proc.returncode == 0:
            # parse output:
            for line in out.splitlines():
                parts = bytestostr(line).split("=")
                if len(parts) != 2:
                    continue
                k = parts[0].strip()
                v = parts[1].strip()
                if k in ("GLX", "GLU.version", "opengl", "pyopengl", "accelerate", "shading-language-version"):
                    props[k] = parse_version(v)
                else:
                    props[k] = v
            log("opengl props=%s", props)
            if props:
                glprops = typedict(props)
                if glprops.strget("success").lower() in TRUE_OPTIONS:
                    log.info(f"OpenGL is supported on display {display_name!r}")
                    renderer = glprops.strget("renderer").split(";")[0]
                    if renderer:
                        log.info(f" using {renderer!r} renderer")
                else:
                    log.info("OpenGL is not supported on this display")
                    probe_err = glprops.strget("error")
                    if probe_err:
                        log.info(f" {probe_err}")
            else:
                log.info("No OpenGL information available")
        else:
            error = bytestostr(err).strip("\n\r")
            for x in str(err).splitlines():
                if x.startswith("RuntimeError: "):
                    error = x[len("RuntimeError: "):]
                    break
                if x.startswith("ImportError: "):
                    error = x[len("ImportError: "):]
                    break
            props["error"] = error
            log.warn("Warning: OpenGL support check failed:")
            log.warn(f" {error}")
    except Exception as e:
        log("query_opengl()", exc_info=True)
        log.error("Error: OpenGL support check failed")
        log.error(f" {e!r}")
        props["error"] = str(e)
    log("OpenGL: %s", props)
    return props


def load_opengl() -> dict[str, Any]:
    with OSEnvContext(XPRA_VERIFY_MAIN_THREAD="0"):
        try:
            # import OpenGL directly
            import OpenGL
            assert OpenGL
            log("found pyopengl version %s", OpenGL.__version__)
            # this may trigger an `AttributeError` if libGLX / libOpenGL are not installed:
            from OpenGL import GL
            assert GL
            log("loaded `GL` bindings: %s", GL)
        except (ImportError, AttributeError) as e:
            return {
                'error': f'OpenGL is not available: {e}',
                'success': False,
            }
        try:
            from xpra.opengl import backing
            assert backing
        except ImportError:
            return {
                'error': '`xpra.opengl` is not available',
                'success': False,
            }
    return {}


class OpenGLInfo(StubServerMixin):
    PREFIX = "opengl"

    def __init__(self):
        StubServerMixin.__init__(self)
        self.display = os.environ.get("DISPLAY", "")
        self.opengl = "no"
        self.opengl_props: dict[str, Any] = {}

    def init(self, opts) -> None:
        self.opengl = opts.opengl

    def threaded_setup(self) -> None:
        self.opengl_props = self.query_opengl()

    def query_opengl(self) -> dict[str, Any]:
        props: dict[str, Any] = {}
        if self.opengl.lower() == "noprobe" or self.opengl.lower() in FALSE_OPTIONS:
            log("query_opengl() skipped because opengl=%s", self.opengl)
            return props
        err = load_opengl()
        if err:
            return err
        from xpra.platform.paths import get_xpra_command
        cmd = self.get_full_child_command(get_xpra_command() + ["opengl", "--opengl=force"])
        return run_opengl_probe(cmd, self.get_child_env(), self.display)

    def get_caps(self, source) -> dict[str, Any]:
        caps: dict[str, Any] = {}
        if FULL_INFO and self.opengl_props:
            caps[OpenGLInfo.PREFIX] = dict_version_trim(self.opengl_props)
        return caps

    def get_info(self, _proto) -> dict[str, Any]:
        info: dict[str, Any] = {}
        if self.opengl_props:
            info = dict(self.opengl_props)
        return {OpenGLInfo.PREFIX: info}



# This file is part of Xpra.
# Copyright (C) 2025 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import string
import random
from time import monotonic
from subprocess import Popen
from collections.abc import Sequence, Callable

from xpra.auth.common import get_exec_env
from xpra.auth.sys_auth_base import SysAuthenticator, log
from xpra.platform.paths import get_nodock_command
from xpra.util.objects import typedict


def gen_secret(mode: str, count: int) -> str:
    if mode in ("digits", "numeric"):
        options = string.digits
    elif mode in ("alpha", "chars", "characters"):
        options = string.ascii_uppercase
    elif mode == "alphanumeric":
        options = string.ascii_uppercase + string.digits
    else:
        raise RuntimeError(f"unsupported mode {mode!r}")
    return "".join(random.choice(options) for _ in range(count))


class Authenticator(SysAuthenticator):

    def __init__(self, **kwargs):
        log("otpscreen.Authenticator(%s)", kwargs)
        self.mode = kwargs.pop("mode", "digits")
        self.count = int(kwargs.pop("count", 6))
        self.timeout = int(kwargs.pop("timeout", 120))
        self.display = kwargs.pop("display", "auto")
        super().__init__(**kwargs)
        self.uid = -1
        self.gid = -1
        self.secret = ""
        self.valid_until = 0.0
        self.authenticate_check: Callable = self.authenticate_hmac
        self.otp_dialog = None

    def get_uid(self) -> int:
        return self.uid

    def get_gid(self) -> int:
        return self.gid

    def requires_challenge(self) -> bool:
        return True

    def get_password(self) -> str:
        return self.secret

    def authenticate_otp(self, caps: typedict) -> bool:
        self.stop_otp_dialog()
        return self.authenticate_hmac(caps)

    def get_challenge(self, digests: Sequence[str]) -> tuple[bytes, str]:
        if self.salt is not None:
            log.error("Error: authentication challenge already sent!")
            return b"", ""
        self.secret = gen_secret(self.mode, self.count)
        log(f"secret is {self.secret!r}")
        self.valid_until = monotonic() + self.timeout
        cmd = get_nodock_command() + ["otp", self.secret, str(self.timeout)]
        env = get_exec_env(self.display)
        log("otp dialog: Popen(%s, %s)", cmd, env)
        self.otp_dialog = Popen(cmd, env=env)
        from xpra.util.child_reaper import get_child_reaper
        get_child_reaper().add_process(self.otp_dialog, "otp-dialog", cmd, ignore=True, forget=True)
        return super().do_get_challenge(digests)

    def cleanup(self):
        self.stop_otp_dialog()

    def stop_otp_dialog(self) -> None:
        proc = self.otp_dialog
        log("stop_otp_dialog() otp_dialog=%s", proc)
        if proc and proc.poll() is None:
            proc.terminate()

    def __repr__(self):
        return "otpscreen"



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
import sys
from typing import Any

from xpra.net.common import Packet
from xpra.net.protocol.socket_handler import SocketProtocol
from xpra.server.subsystem.stub import StubServerMixin
from xpra.common import FULL_INFO
from xpra.os_util import get_machine_id, gi_import, get_user_uuid
from xpra.util.version import XPRA_NUMERIC_VERSION
from xpra.util.objects import typedict
from xpra.log import Logger

# pylint: disable=import-outside-toplevel

GLib = gi_import("GLib")

log = Logger("server")


# noinspection PyMethodMayBeStatic
class IDServer(StubServerMixin):
    """
    Servers that expose info data via info request.
    """

    def __init__(self):
        self.hello_request_handlers["id"] = self._handle_hello_request_id
        self.uuid = ""

    def get_caps(self, source):
        caps = {}
        if source is None or "versions" in source.wants:
            caps["uuid"] = get_user_uuid()
            mid = get_machine_id()
            if mid:
                caps["machine_id"] = mid
        return caps

    def _handle_hello_request_id(self, proto, _caps: typedict) -> bool:
        self.send_id_info(proto)
        return True

    def send_id_info(self, proto: SocketProtocol) -> None:
        log("id info request from %s", proto._conn)
        proto._log_stats = False
        proto.send_now(Packet("hello", self.get_session_id_info()))

    def get_session_id_info(self) -> dict[str, Any]:
        # minimal information for identifying the session
        id_info = {
            "session-type": self.session_type,
            "session-name": self.session_name,
            "uuid": self.uuid,
            "platform": sys.platform,
            "pid": os.getpid(),
            "machine-id": get_machine_id(),
            "version": XPRA_NUMERIC_VERSION[:FULL_INFO+1],
        }
        display = os.environ.get("DISPLAY", "")
        if display:
            id_info["display"] = display
        return id_info



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os.path
from urllib.parse import unquote
from collections.abc import Callable

from xpra.net.file_transfer import FileTransferHandler
from xpra.client.gtk3.window.stub_window import GtkStubWindow
from xpra.os_util import gi_import, WIN32
from xpra.util.objects import typedict
from xpra.util.str_fn import csv
from xpra.log import Logger

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
Gio = gi_import("Gio")

log = Logger("window", "events", "dragndrop")


def xid(w) -> int:
    # TODO: use a generic window handle function
    # this only used for debugging for now
    if w:
        try:
            return w.get_xid()
        except AttributeError:
            pass
    return 0


def drag_drop_cb(widget, context, x: int, y: int, time: int) -> None:
    targets = list(x.name() for x in context.list_targets())
    log("drag_drop_cb%s targets=%s", (widget, context, x, y, time), targets)
    if not targets:
        # this happens on macOS, but we can still get the data...
        log("Warning: no targets provided, continuing anyway")
    elif "text/uri-list" not in targets:
        log("Warning: cannot handle targets:")
        log(" %s", csv(targets))
        return
    atom = Gdk.Atom.intern("text/uri-list", False)
    widget.drag_get_data(context, atom, time)


def drag_motion_cb(wid: int, context, x: int, y: int, time: int) -> bool:
    log("drag_motion_cb(%#x, %s, %i, %i, %i)", wid, context, x, y, time)
    Gdk.drag_status(context, Gdk.DragAction.COPY, time)
    return True  # accept this data


class DragNDropWindow(GtkStubWindow):

    def init_window(self, client, metadata: typedict, client_props: typedict) -> None:
        # opengl probing uses a fake client,
        # in which case we don't want dragndrop initialized:
        if isinstance(client, FileTransferHandler):
            self._file_handler = client
            self.init_dragndrop()

    def is_readonly(self) -> bool:
        # inject dependency on UI client,
        # the readonly flag can be changed from the tray menu:
        return getattr(self, "readonly", False)

    ######################################################################
    # drag and drop:
    def init_dragndrop(self) -> None:
        targets = [
            Gtk.TargetEntry.new("text/uri-list", 0, 80),
        ]
        flags = Gtk.DestDefaults.MOTION | Gtk.DestDefaults.HIGHLIGHT
        actions = Gdk.DragAction.COPY  # | Gdk.ACTION_LINK
        self.drag_dest_set(flags, targets, actions)
        self.connect("drag_drop", drag_drop_cb)
        self.connect("drag_motion", drag_motion_cb)
        self.connect("drag_data_received", self.drag_got_data_cb)

    def drag_got_data_cb(self, wid: int, context, x: int, y: int, selection, info, time: int) -> None:
        log("drag_got_data_cb(%#x, %s, %i, %i, %r, %s, %i)", wid, context, x, y, selection, info, time)
        if self.is_readonly():
            return
        targets = list(x.name() for x in context.list_targets())
        actions = context.get_actions()

        dest_window = xid(context.get_dest_window())
        source_window = xid(context.get_source_window())
        suggested_action = context.get_suggested_action()
        log("drag_got_data_cb context: source_window=%#x, dest_window=%#x", source_window, dest_window)
        log(f"drag_got_data_cb context: {suggested_action=}, {actions=}, {targets=}")
        dtype = selection.get_data_type()
        fmt = selection.get_format()
        length = selection.get_length()
        target = selection.get_target()
        text = selection.get_text()
        uris = selection.get_uris()
        log("drag_got_data_cb selection: data type=%s, format=%s, length=%s, target=%s, text=%s, uris=%s",
            dtype, fmt, length, target, text, uris)
        if not uris:
            return
        filelist = []
        for uri in uris:
            if not uri:
                continue
            if not uri.startswith("file://"):
                log.warn("Warning: cannot handle drag-n-drop URI '%s'", uri)
                continue
            filename = unquote(uri[len("file://"):].rstrip("\n\r"))
            if WIN32:
                filename = filename.lstrip("/")
            abspath = os.path.abspath(filename)
            if not os.path.isfile(abspath):
                log.warn("Warning: '%s' is not a file", abspath)
                continue
            filelist.append(abspath)
        log("drag_got_data_cb: will try to upload: %s", csv(filelist))
        pending = set(filelist)

        def file_done(filename: str) -> None:
            if not pending:
                return
            try:
                pending.remove(filename)
            except KeyError:
                pass
            # when all the files have been loaded / failed,
            # finish the drag and drop context so the source knows we're done with them:
            if not pending:
                context.finish(True, False, time)

        # we may want to only process a limited number of files "at the same time":
        for filename in filelist:
            self.drag_process_file(filename, file_done)

    def drag_process_file(self, filename: str, file_done_cb: Callable) -> None:
        if self.is_readonly():
            return

        def got_file_info(gfile, result, arg=None):
            log("got_file_info(%s, %s, %s)", gfile, result, arg)
            file_info = gfile.query_info_finish(result)
            basename = gfile.get_basename()
            ctype = file_info.get_content_type()
            size = file_info.get_size()
            log("file_info(%s)=%s ctype=%s, size=%s", filename, file_info, ctype, size)

            def got_file_data(gfile, result, user_data=None) -> None:
                _, data, entity = gfile.load_contents_finish(result)
                filesize = len(data)
                log("got_file_data(%s, %s, %s) entity=%s", gfile, result, user_data, entity)
                file_done_cb(filename)
                openit = self._file_handler.remote_open_files
                log.info("sending file %s (%i bytes)", basename, filesize)
                self._file_handler.send_file(filename, "", data, filesize=filesize, openit=openit)

            cancellable = None
            user_data = (filename, True)
            gfile.load_contents_async(cancellable, got_file_data, user_data)

        try:
            gfile = Gio.File.new_for_path(path=filename)
            # basename = gf.get_basename()
            FILE_QUERY_INFO_NONE = 0
            G_PRIORITY_DEFAULT = 0
            cancellable = None
            gfile.query_info_async("standard::*", FILE_QUERY_INFO_NONE, G_PRIORITY_DEFAULT,
                                   cancellable, got_file_info, None)
        except Exception as e:
            log("file upload for %s:", filename, exc_info=True)
            log.error("Error: cannot upload '%s':", filename)
            log.estr(e)
            del e
            file_done_cb(filename)



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import os
from typing import Any

from xpra.os_util import gi_import
from xpra.log import Logger

metalog = Logger("metadata")

GObject = gi_import("GObject")

PROPERTIES_DEBUG = [x.strip() for x in os.environ.get("XPRA_WINDOW_PROPERTIES_DEBUG", "").split(",")]


def n_arg_signal(n: int) -> tuple:
    return GObject.SignalFlags.RUN_LAST, GObject.TYPE_NONE, (GObject.TYPE_PYOBJECT,) * n


no_arg_signal = n_arg_signal(0)
one_arg_signal = n_arg_signal(1)


def to_gsignals(signals: dict[str, Any]) -> dict[str, tuple]:
    gsignals: dict[str, tuple] = {}
    for key, value in signals.items():
        if isinstance(value, tuple):
            gsignals[key] = value
        elif isinstance(value, int):
            gsignals[key] = n_arg_signal(value)
        else:
            raise ValueError("unexpected signal value: %r (%s)" % (value, type(value)))
    return gsignals


class AutoPropGObjectMixin:
    """Mixin for automagic property support in GObjects.

    Make sure this is the first entry on your parent list, so super().__init__
    will work right."""

    def __init__(self):
        self._gproperties: dict[str, Any] = {}

    def do_get_property(self, pspec):
        return self._gproperties.get(pspec.name)

    def do_set_property(self, pspec, value) -> None:
        self._internal_set_property(pspec.name, value)

    def _internal_set_property(self, name: str, value) -> None:
        if name in PROPERTIES_DEBUG:
            metalog.info("_internal_set_property(%r, %r)", name, value, backtrace=True)
        self._gproperties[name] = value
        self.notify(name)



# This file is part of Xpra.
# Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.client.gtk3.window.base import HAS_X11_BINDINGS
from xpra.client.gtk3.window.factory import get_window_base_classes
from xpra.gtk.util import event_mask_strs
from xpra.util.objects import typedict
from xpra.os_util import gi_import
from xpra.util.str_fn import bytestostr
from xpra.log import Logger

log = Logger("window")
paintlog = Logger("paint")
metalog = Logger("metadata")
geomlog = Logger("geometry")

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GdkPixbuf = gi_import("GdkPixbuf")
GObject = gi_import("GObject")
Gio = gi_import("Gio")

WINDOW_BASES = get_window_base_classes()
WindowBaseClass = type("WindowBaseClass", WINDOW_BASES, {})
log(f"WindowBaseClass: {WINDOW_BASES=}")


def get_all_gsignals() -> dict[str, Any]:
    all_gsignals = {}
    for bc in WINDOW_BASES:
        gsignals = getattr(bc, "__gsignals__", {})
        all_gsignals.update(gsignals)
    log(f"gsignals: {all_gsignals.keys()}")
    return all_gsignals


class ClientWindow(WindowBaseClass):
    """
    GTK3 version of the ClientWindow class
    """
    __gsignals__ = get_all_gsignals()

    def init_window(self, client, metadata: typedict, client_props: typedict) -> None:
        self.menu_helper = None
        for bc in WINDOW_BASES:
            bc.init_window(self, client, metadata, client_props)

    def destroy(self) -> None:  # pylint: disable=method-hidden
        for bc in WINDOW_BASES:
            bc.cleanup(self)
        super().destroy()

    def get_info(self) -> dict[str, Any]:
        info: dict[str, Any] = {}
        for bc in WINDOW_BASES:
            info.update(bc.get_info(self))
        return info

    def get_window_event_mask(self) -> Gdk.EventMask:
        mask = 0
        for bc in WINDOW_BASES:
            if hasattr(bc, "get_window_event_mask"):
                mask |= bc.get_window_event_mask(self)
            log("%s.get_window_event_mask()=%s", bc, event_mask_strs(mask))
        log("get_window_event_mask()=%s", event_mask_strs(mask))
        return mask

    def init_widget_events(self, widget) -> None:
        for bc in WINDOW_BASES:
            if hasattr(bc, "init_widget_events"):
                bc.init_widget_events(self, widget)

    def set_icon(self, pixbuf: GdkPixbuf.Pixbuf) -> None:
        for bc in WINDOW_BASES:
            if hasattr(bc, "set_icon"):
                bc.set_icon(self, pixbuf)

    def show_xpra_menu(self, *_args) -> None:
        # this is called by the headerbar
        mh = self._client.get_window_menu_helper()
        if mh:
            mh.build()
            mh.popup(0, 0)

    def show_window_menu(self, *_args) -> None:
        if not self.menu_helper:
            from xpra.client.gtk3.window.menu import WindowMenuHelper
            self.menu_helper = WindowMenuHelper(self._client, self)
            self.menu_helper.build()
        self.menu_helper.popup(0, 0)

    def get_backing_class(self) -> type:
        from xpra.client.gtk3.cairo_backing import CairoBacking
        return CairoBacking

    def xget_u32_property(self, target, name: str, default_value=0) -> int:
        if HAS_X11_BINDINGS:
            from xpra.client.gtk3.window.base import GTKClientWindowBase
            return GTKClientWindowBase.xget_u32_property(self, target, name, default_value)
        # pure Gdk lookup:
        try:
            name_atom = Gdk.Atom.intern(name, False)
            type_atom = Gdk.Atom.intern("CARDINAL", False)
            prop = Gdk.property_get(target, name_atom, type_atom, 0, 9999, False)
            if not prop or len(prop) != 3 or len(prop[2]) != 1:
                return default_value
            metalog("xget_u32_property(%s, %s, %s)=%s", target, name, default_value, prop[2][0])
            return prop[2][0]
        except Exception as e:
            metalog.error("xget_u32_property error on %s / %s: %s", target, name, e)
            return default_value

    def get_drawing_area_geometry(self) -> tuple[int, int, int, int]:
        gdkwindow = self.drawing_area.get_window()
        if gdkwindow:
            x, y = gdkwindow.get_origin()[1:]
        else:
            x, y = self.get_position()
        w, h = self.get_size()
        return x, y, w, h

    def apply_geometry_hints(self, hints: typedict) -> None:
        """ we convert the hints as a dict into a gdk.Geometry + gdk.WindowHints """
        wh = Gdk.WindowHints
        name_to_hint = {
            "max_width": wh.MAX_SIZE,
            "max_height": wh.MAX_SIZE,
            "min_width": wh.MIN_SIZE,
            "min_height": wh.MIN_SIZE,
            "base_width": wh.BASE_SIZE,
            "base_height": wh.BASE_SIZE,
            "width_inc": wh.RESIZE_INC,
            "height_inc": wh.RESIZE_INC,
            "min_aspect_ratio": wh.ASPECT,
            "max_aspect_ratio": wh.ASPECT,
        }
        # these fields can be copied directly to the gdk.Geometry as ints:
        INT_FIELDS: list[str] = [
            "min_width", "min_height",
            "max_width", "max_height",
            "base_width", "base_height",
            "width_inc", "height_inc",
        ]
        ASPECT_FIELDS: dict[str, str] = {
            "min_aspect_ratio": "min_aspect",
            "max_aspect_ratio": "max_aspect",
        }
        thints = typedict(hints)
        if self.drawing_area:
            # apply min size to the drawing_area:
            # (for CSD mode, ie: headerbar)
            minw = thints.intget("min_width", 0)
            minh = thints.intget("min_height", 0)
            self.drawing_area.set_size_request(minw, minh)

        geom = Gdk.Geometry()
        mask = 0
        for k, v in hints.items():
            k = bytestostr(k)
            if k in INT_FIELDS:
                setattr(geom, k, v)
                mask |= int(name_to_hint.get(k, 0))
            elif k in ASPECT_FIELDS:
                field = ASPECT_FIELDS[k]
                setattr(geom, field, float(v))
                mask |= int(name_to_hint.get(k, 0))
        gdk_hints = Gdk.WindowHints(mask)
        geomlog("apply_geometry_hints(%s) geometry=%s, hints=%s", hints, geom, gdk_hints)
        self.set_geometry_hints(self.drawing_area, geom, gdk_hints)

    def can_maximize(self) -> bool:
        hints = self.geometry_hints
        if not hints:
            return True
        maxw = hints.intget("max_width", 32768)
        maxh = hints.intget("max_height", 32768)
        if maxw > 32000 and maxh > 32000:
            return True
        geom = self.get_drawing_area_geometry()
        dw, dh = geom[2], geom[3]
        return dw < maxw and dh < maxh

    def draw_widget(self, widget, context) -> bool:
        paintlog("draw_widget(%s, %s)", widget, context)
        if not self.get_mapped():
            return False
        backing = self._backing
        if not backing:
            return False
        w, h = self.get_size()
        backing.cairo_draw(context, w, h)
        return True

    def get_map_client_properties(self) -> dict[str, Any]:
        props = {}
        for bc in WINDOW_BASES:
            props.update(bc.get_map_client_properties(self))
        return props

    def get_configure_client_properties(self) -> dict[str, Any]:
        props = {}
        for bc in WINDOW_BASES:
            props.update(bc.get_configure_client_properties(self))
        return props


GObject.type_register(ClientWindow)



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import math
from collections.abc import Sequence

from xpra.os_util import gi_import
from xpra.util.str_fn import memoryview_to_bytes
from xpra.x11.bindings.core import get_root_xid
from xpra.x11.bindings.window import X11WindowBindings
from xpra.x11.bindings.fixes import XFixesBindings
from xpra.x11.bindings.composite import XCompositeBindings
from xpra.x11.error import xsync, xswallow, xlog
from xpra.codecs.argb.argb import unpremultiply_argb, bgra_to_rgba, bgra_to_rgbx, r210_to_rgbx, bgr565_to_rgbx
from cairo import OPERATOR_OVER, OPERATOR_SOURCE  # pylint: disable=no-name-in-module
from xpra.log import Logger

GLib = gi_import("GLib")
GObject = gi_import("GObject")
Gdk = gi_import("Gdk")
GdkX11 = gi_import("GdkX11")

log = Logger("screen")

XComposite = XCompositeBindings()
X11Window = X11WindowBindings()
XFixes = XFixesBindings()


def fill_rect(cr, shade: tuple[float, float, float], x: int, y: int, w: int, h: int) -> None:
    log("fill_rect%s", (shade, x, y, w, h))
    cr.new_path()
    cr.set_source_rgb(*shade)
    cr.rectangle(x, y, w, h)
    cr.fill()


def draw_rect(cr, shade: tuple[float, float, float], x: int, y: int, w: int, h: int) -> None:
    log("draw_rect%s", (shade, x, y, w, h))
    cr.new_path()
    cr.set_line_width(2)
    cr.set_source_rgb(*shade)
    cr.rectangle(x, y, w, h)
    cr.stroke()


def init_root_overlay() -> int:
    try:
        import cairo
        log(f"init_root_overlay() found cairo: {cairo}")
        with xsync:
            rxid = get_root_xid()
            root_overlay = XComposite.XCompositeGetOverlayWindow(rxid)
            log("init_root_overlay() root_overlay=%#x", root_overlay)
            if root_overlay:
                from xpra.x11.prop import prop_set
                prop_set(root_overlay, "WM_TITLE", "latin1", "RootOverlay")
                XFixes.AllowInputPassthrough(root_overlay)
                log("init_root_overlay() done AllowInputPassthrough(%#x)", root_overlay)
                return root_overlay
    except Exception as e:
        log("XCompositeGetOverlayWindow(%#x)", rxid, exc_info=True)
        log.error("Error setting up xvfb synchronization:")
        log.estr(e)
        release_root_overlay(root_overlay)
    return 0


def release_root_overlay(xid: int) -> None:
    with xswallow:
        XComposite.XCompositeReleaseOverlayWindow(xid)


def cairo_create(root_overlay: int):
    display = Gdk.Display.get_default()
    overlaywin = GdkX11.X11Window.foreign_new_for_display(display, root_overlay)
    return overlaywin.cairo_create()


def paint_overlay_monitors(cr, screen: Sequence) -> None:
    log("paint_overlay_monitors%s", (cr, screen))
    if len(screen) < 10:
        return
    display_name, width, height, width_mm, height_mm, monitors, work_x, work_y, work_width, work_height = screen[:10]
    assert display_name or width_mm or height_mm or True  # just silences pydev warnings
    # paint dark grey background for display dimensions:
    fill_rect(cr, (0.2, 0.2, 0.2), 0, 0, width, height)
    fill_rect(cr, (0.2, 0.2, 0.4), 0, 0, width, height)
    # paint lighter grey background for workspace dimensions:
    draw_rect(cr, (0.5, 0.5, 0.5), work_x, work_y, work_width, work_height)
    # paint each monitor with even lighter shades of grey:
    for m in monitors:
        if len(m) < 7:
            continue
        plug_name, plug_x, plug_y, plug_width, plug_height, plug_width_mm, plug_height_mm = m[:7]
        assert plug_name or plug_width_mm or plug_height_mm or True  # just silences pydev warnings
        draw_rect(cr, (0.7, 0.7, 0.7), plug_x, plug_y, plug_width, plug_height)
        if len(m) >= 10:
            dwork_x, dwork_y, dwork_width, dwork_height = m[7:11]
            draw_rect(cr, (1, 1, 1), dwork_x, dwork_y, dwork_width, dwork_height)


def paint_root_overlay_windows(cr, windows: Sequence) -> None:
    log("paint_root_overlay_windows(%s)",windows)
    for window in windows:
        w, h = window.get_property("geometry")[2:4]
        image = window.get_image(0, 0, w, h)
        if not image:
            continue
        paint_overlay_window(cr, window, 0, 0, image)
        paint_overlay_frame(cr, window)


def update_root_overlay(root_overlay: int, window, x: int, y: int, image) -> None:
    log("update_root_overlay%s", (root_overlay, window, x, y, image))
    display = Gdk.Display.get_default()
    overlaywin = GdkX11.X11Window.foreign_new_for_display(display, root_overlay)
    cr = overlaywin.cairo_create()
    paint_overlay_window(cr, window, x, y, image)


def paint_overlay_window(cr, window, x: int, y: int, image) -> None:
    log("paint_overlay_window%s", (cr, window, x, y, image))
    wx, wy = window.get_property("geometry")[:2]
    # FIXME: we should paint the root overlay directly
    # either using XCopyArea or XShmPutImage,
    # using GTK and having to unpremultiply then convert to RGB is just too slooooow
    width = image.get_width()
    height = image.get_height()
    rowstride = image.get_rowstride()
    img_data = image.get_pixels()
    rgb_format = image.get_pixel_format()
    log("update_root_overlay%s rgb_format=%s, img_data=%i (%s)",
        (window, x, y, image), rgb_format, len(img_data), type(img_data))
    operator = OPERATOR_SOURCE
    if rgb_format == "BGRA":
        img_data = unpremultiply_argb(img_data)
        img_data = bgra_to_rgba(img_data)
        operator = OPERATOR_OVER
    elif rgb_format == "BGRX":
        img_data = bgra_to_rgbx(img_data)
    elif rgb_format == "r210":
        # lossy...
        img_data = r210_to_rgbx(img_data, width, height, rowstride, width * 4)
        rowstride = width * 4
    elif rgb_format == "BGR565":
        img_data = bgr565_to_rgbx(img_data)
        rowstride *= 2
    else:
        raise ValueError(f"xync-xvfb root overlay paint code does not handle {rgb_format} pixel format")
    img_data = memoryview_to_bytes(img_data)
    log("update_root_overlay%s painting rectangle %s", (window, x, y, image), (wx + x, wy + y, width, height))
    from xpra.gtk.pixbuf import get_pixbuf_from_data
    pixbuf = get_pixbuf_from_data(img_data, True, width, height, rowstride)
    cr.new_path()
    cr.rectangle(wx + x, wy + y, width, height)
    cr.clip()
    Gdk.cairo_set_source_pixbuf(cr, pixbuf, wx + x, wy + y)
    cr.set_operator(operator)
    cr.paint()
    with xlog:
        image.free()


def paint_overlay_frame(cr, window) -> None:
    x, y, w, h = window.get_property("geometry")[:4]
    frame = window.get_property("frame")
    log("paint_overlay_frame(%s, %s) frame=%s", cr, window, frame)
    if frame and tuple(frame) != (0, 0, 0, 0):
        left, right, top, bottom = frame
        # always add a little something, so we can see the edge:
        left = max(1, left)
        right = max(1, right)
        top = max(1, top)
        bottom = max(1, bottom)
        rectangles = (
            (x - left, y, left, h, True),  # left side
            (x - left, y - top, w + left + right, top, True),  # top
            (x + w, y, right, h, True),  # right
            (x - left, y + h, w + left + right, bottom, True),  # bottom
        )
    else:
        rectangles = (
            (x, y, w, h, False),
        )
    log("rectangles for window frame=%s and geometry=%s : %s", frame, (x, y, w, h), rectangles)
    for x, y, w, h, fill in rectangles:
        cr.new_path()
        cr.set_source_rgb(0.1, 0.1, 0.1)
        cr.set_line_width(1)
        cr.rectangle(x, y, w, h)
        if fill:
            cr.fill()
        else:
            cr.stroke()


def paint_overlay_pointer(cr, x: int, y: int) -> None:
    log("paint_overlay_pointer(%s, %i, %i)", cr, x, y)
    cr.set_source_rgb(1.0, 0.5, 0.7)
    cr.new_path()
    cr.arc(x, y, 10.0, 0, 2.0 * math.pi)
    cr.stroke_preserve()
    cr.set_source_rgb(0.3, 0.4, 0.6)
    cr.fill()



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2017 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

import sys

from xpra.exit_codes import ExitCode
from xpra.gtk.css_overrides import add_screen_css
from xpra.gtk.window import add_close_accel
from xpra.os_util import gi_import
from xpra.log import Logger
from xpra.util.env import envbool

Gtk = gi_import("Gtk")
Gdk = gi_import("Gdk")
GLib = gi_import("GLib")

log = Logger("auth")

SHOW = envbool("XPRA_OTP_SHOW", True)

CSS = b"""
.otp {
    font-family: 'monospace';
    font-size: 36px;
    padding: 12px 18px;
    border-radius: 10px;
    background: rgba(255,255,255,0.03);
    box-shadow: 0 2px 8px rgba(0,0,0,0.12);
}

.muted {
    color: rgba(255,255,255,0.7);
}

window {
    background-image: linear-gradient(120deg, #0f172a, #0b1220);
    color: #e6eef8;
}

button {
    border-radius: 8px;
    padding: 6px 10px;
}
"""


class OTPDialog(Gtk.Window):
    def __init__(self, otp: str, lifetime: int = 30):
        super().__init__(title="One-Time Password")
        self.set_default_size(420, 200)
        self.set_border_width(12)
        self.lifetime = lifetime
        self.remaining = lifetime
        self.otp = otp
        self.revealed = SHOW

        add_screen_css(CSS)
        add_close_accel(self, Gtk.main_quit)

        main_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL, spacing=12)
        self.add(main_box)

        # Top title and subtitle
        title = Gtk.Label()
        title.set_markup("<span size='xx-large' weight='bold'>Authentication code</span>")
        title.set_halign(Gtk.Align.START)
        main_box.pack_start(title, False, False, 0)

        subtitle = Gtk.Label(label="Use this code to complete your sign-in.")
        subtitle.set_halign(Gtk.Align.START)
        subtitle.get_style_context().add_class("muted")
        main_box.pack_start(subtitle, False, False, 0)

        # OTP display area
        otp_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)
        otp_box.set_halign(Gtk.Align.CENTER)
        main_box.pack_start(otp_box, True, True, 0)

        self.otp_label = Gtk.Label()
        self.otp_label.set_markup(self._masked_otp(self.otp))
        self.otp_label.get_style_context().add_class("otp")
        self.otp_label.set_selectable(True)
        otp_box.pack_start(self.otp_label, True, True, 20)

        controls_box = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=6)
        otp_box.pack_start(controls_box, False, False, 0)

        # Reveal/hide toggle
        self.reveal_button = Gtk.Button()
        self._set_reveal_icon()
        self.reveal_button.set_tooltip_text("Show code")
        self.reveal_button.connect("clicked", self.on_reveal_clicked)
        self.reveal_button.set_size_request(48, 48)
        controls_box.pack_start(self.reveal_button, False, False, 0)

        # Copy button
        copy_button = Gtk.Button.new_from_icon_name("edit-copy", Gtk.IconSize.BUTTON)
        copy_button.set_tooltip_text("Copy code to clipboard")
        copy_button.connect("clicked", self.on_copy_clicked)
        copy_button.set_size_request(48, 48)
        controls_box.pack_start(copy_button, False, False, 0)

        # Progress + time remaining
        self.progress = Gtk.ProgressBar(show_text=True)
        self.progress.set_fraction(1.0)
        main_box.pack_start(self.progress, False, False, 0)

        # Footer actions
        footer = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=8)
        footer.set_halign(Gtk.Align.END)
        main_box.pack_start(footer, False, False, 0)

        close_btn = Gtk.Button(label="Close")
        close_btn.connect("clicked", lambda w: self.close())
        footer.pack_start(close_btn, False, False, 0)

        # Start timer
        GLib.timeout_add_seconds(1, self._tick)

        self.connect("destroy", Gtk.main_quit)
        self.show_all()

    def _masked_otp(self, otp: str) -> str:
        if self.revealed:
            display = otp
        else:
            display = "•" * len(otp)
        # markup to increase tracking/letter spacing
        return f"<span font='Monospace 28'>{display}</span>"

    def _set_reveal_icon(self):
        icon_name = "view-visible" if not self.revealed else "view-hidden"
        # fallback to stock label if icon isn't available
        if Gtk.IconTheme.get_default().has_icon(icon_name):
            image = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.BUTTON)
            self.reveal_button.set_image(image)
            self.reveal_button.set_always_show_image(True)
        else:
            self.reveal_button.set_label("Show" if not self.revealed else "Hide")

    def on_reveal_clicked(self, _):
        self.revealed = not self.revealed
        self._set_reveal_icon()
        self.otp_label.set_markup(self._masked_otp(self.otp))
        self.reveal_button.set_tooltip_text("Hide code" if self.revealed else "Show code")

    def on_copy_clicked(self, _):
        clipboard = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
        clipboard.set_text(self.otp, -1)
        clipboard.store()
        # provide visual feedback briefly
        prev_label = self.reveal_button.get_tooltip_text()
        self.reveal_button.set_tooltip_text("Copied!")
        GLib.timeout_add_seconds(2, lambda: (self.reveal_button.set_tooltip_text(prev_label) or False))

    def _tick(self) -> bool:
        self.remaining -= 1
        if self.remaining < 0:
            self.progress.set_fraction(0.0)
            self.progress.set_text("Expired")
            self.otp_label.set_markup("<span font='Monospace 20'>— expired —</span>")
            return False  # stop ticking

        fraction = self.remaining / float(self.lifetime)
        self.progress.set_fraction(fraction)
        self.progress.set_text(f"{self.remaining}s remaining")
        # subtly update otp label color if close to expiry
        if self.remaining <= 5:
            ctx = self.otp_label.get_style_context()
            ctx.add_class("muted")
        return True


def main(args: list[str]) -> int:
    if not args:
        return ExitCode.NO_DATA
    otp = args[0]
    try:
        lifetime = int(args[1])
    except (ValueError, IndexError):
        lifetime = 30

    win = OTPDialog(otp=otp, lifetime=lifetime)
    win.set_position(Gtk.WindowPosition.CENTER)
    win.show_all()
    Gtk.main()
    return ExitCode.OK


if __name__ == "__main__":
    r = main(sys.argv[1:])
    sys.exit(r)



#!/usr/bin/env python3
# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from collections.abc import Callable

from xpra.platform import platform_import


def get_backends() -> list[Callable]:
    return []


platform_import(globals(), "notification", False, "get_backends")



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Any

from xpra.common import noop
from xpra.os_util import gi_import, WIN32, OSX
from xpra.util.system import is_X11
from xpra.util.objects import typedict
from xpra.util.env import envint, envbool
from xpra.client.gtk3.window.stub_window import GtkStubWindow
from xpra.util.gobject import one_arg_signal
from xpra.gtk.util import ds_inited
from xpra.log import Logger

GLib = gi_import("GLib")
Gdk = gi_import("Gdk")

log = Logger("focus", "grab")

NotifyInferior = 2

FOCUS_RECHECK_DELAY = envint("XPRA_FOCUS_RECHECK_DELAY", 0)
AUTOGRAB_WITH_FOCUS = envbool("XPRA_AUTOGRAB_WITH_FOCUS", False)


class FocusWindow(GtkStubWindow):
    __gsignals__ = {
        "x11-focus-out-event": one_arg_signal,
        "x11-focus-in-event": one_arg_signal,
    }

    def init_window(self, client, metadata: typedict, client_props: typedict) -> None:
        self._focus_latest = None
        self.recheck_focus_timer: int = 0
        self.remove_event_receiver = noop
        self.init_focus()

    def cleanup(self):
        self.remove_event_receiver()
        self.cancel_focus_timer()
        if self._client.has_focus(self.wid):
            self._unfocus()

    def get_info(self) -> dict[str, Any]:
        return {
            "focused": bool(self._focus_latest),
        }

    def get_window_event_mask(self) -> Gdk.EventMask:
        return Gdk.EventMask.FOCUS_CHANGE_MASK

    def init_focus(self) -> None:
        self.when_realized("init-focus", self.do_init_focus)

    def do_init_focus(self) -> None:
        # hook up the X11 gdk event notifications,
        # so we can get focus-out when grabs are active:
        if not (WIN32 or OSX) and is_X11():
            try:
                from xpra.x11.dispatch import add_event_receiver, remove_event_receiver
            except ImportError as e:
                log("do_init_focus()", exc_info=True)
                if not ds_inited():
                    log.warn("Warning: missing Gdk X11 bindings:")
                    log.warn(" %s", e)
                    log.warn(" you may experience window focus issues")
            else:
                log("adding event receiver so we can get FocusIn and FocusOut events whilst grabbing the keyboard")
                xid = self.get_window().get_xid()
                add_event_receiver(xid, self)

                def remove_hook() -> None:
                    remove_event_receiver(xid, self)
                self.remove_event_receiver = remove_hook

        # other platforms should be getting regular focus events instead:

        def focus_in(_window, event) -> None:
            log("focus-in-event for wid=%#x", self.wid)
            self.do_x11_focus_in_event(event)

        def focus_out(_window, event) -> None:
            log("focus-out-event for wid=%#x", self.wid)
            self.do_x11_focus_out_event(event)

        self.connect("focus-in-event", focus_in)
        self.connect("focus-out-event", focus_out)
        if not self._override_redirect:
            self.connect("notify::has-toplevel-focus", self._focus_change)

    def _focus_change(self, *args) -> None:
        assert not self._override_redirect
        htf = self.has_toplevel_focus()
        log("%s focus_change%s has-toplevel-focus=%s, _been_mapped=%s", self, args, htf, self._been_mapped)
        if self._been_mapped:
            self._focus_latest = htf
            self.schedule_recheck_focus()

    def recheck_focus(self) -> None:
        self.recheck_focus_timer = 0
        self.send_latest_focus()

    def send_latest_focus(self) -> None:
        focused = self._client._focused
        log("send_latest_focus() wid=%#x, focused=%s, latest=%s", self.wid, focused, self._focus_latest)
        if self._focus_latest:
            self._focus()
        else:
            self._unfocus()

    def _focus(self) -> bool:
        change = self._client.update_focus(self.wid, True)
        if change and AUTOGRAB_WITH_FOCUS:
            # soft dependency on GrabWindow:
            autograb = getattr(self, "may_autograb", noop)
            autograb()
        return change

    def _unfocus(self) -> bool:
        client = self._client
        client.window_ungrab()
        if client.pointer_grabbed and client.pointer_grabbed == self.wid:
            # we lost focus, assume we also lost the grab:
            client.pointer_grabbed = None
        changed = super()._unfocus()
        if changed and AUTOGRAB_WITH_FOCUS:
            # soft dependency on GrabWindow:
            ungrab = getattr(self, "keyboard_ungrab", noop)
            ungrab()
        return changed

    def cancel_focus_timer(self) -> None:
        rft = self.recheck_focus_timer
        if rft:
            self.recheck_focus_timer = 0
            GLib.source_remove(rft)

    def schedule_recheck_focus(self) -> None:
        if self._override_redirect:
            # never send focus events for OR windows
            return
        # we receive pairs of FocusOut + FocusIn following a keyboard grab,
        # so we recheck the focus status via this timer to skip unnecessary churn
        if FOCUS_RECHECK_DELAY < 0:
            self.recheck_focus()
        elif not self.recheck_focus_timer:
            log(f"will recheck focus in {FOCUS_RECHECK_DELAY}ms")
            self.recheck_focus_timer = GLib.timeout_add(FOCUS_RECHECK_DELAY, self.recheck_focus)

    def do_x11_focus_out_event(self, event) -> None:
        log("do_x11_focus_out_event(%s)", event)
        # `detail` is only available with X11 clients,
        # but this method can be called by other clients (see `focus_out`):
        detail = getattr(event, "detail", 0)
        if detail == NotifyInferior:
            log("dropped NotifyInferior focus event")
            return
        self._focus_latest = False
        self.schedule_recheck_focus()

    def do_x11_focus_in_event(self, event) -> None:
        log("do_x11_focus_in_event(%s) been_mapped=%s", event, self._been_mapped)
        if self._been_mapped:
            self._focus_latest = True
            self.schedule_recheck_focus()



# This file is part of Xpra.
# Copyright (C) 2012 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from typing import Protocol, Any

from xpra.codecs.image import ImageWrapper
from xpra.util.objects import typedict


class VideoCodec(Protocol):
    def init_context(self, encoding: str, width: int, height: int, src_format: str, options: typedict) -> None:
        raise NotImplementedError()

    def get_type(self) -> str:
        raise NotImplementedError()

    def get_info(self) -> dict[str, Any]:
        raise NotImplementedError()

    def get_encoding(self) -> str:
        raise NotImplementedError()

    def get_width(self) -> int:
        raise NotImplementedError()

    def get_height(self) -> int:
        raise NotImplementedError()

    def get_src_format(self) -> str:
        raise NotImplementedError()

    def clean(self) -> None:
        raise NotImplementedError()


class VideoEncoder(VideoCodec):
    def is_closed(self) -> bool:
        raise NotImplementedError()

    def compress_image(self, image: ImageWrapper, options: typedict) -> tuple[bytes, dict]:
        raise NotImplementedError()


class VideoDecoder(VideoCodec):

    def get_colorspace(self) -> str:
        raise NotImplementedError()

    def is_closed(self) -> bool:
        raise NotImplementedError()

    def decompress_image(self, data: bytes, options: typedict) -> ImageWrapper:
        raise NotImplementedError()


class ColorspaceConverter(Protocol):

    def init_context(self, src_width: int, src_height: int, src_format: str,
                     dst_width: int, dst_height: int, dst_format: str, options: typedict) -> None:
        raise NotImplementedError()

    def get_type(self) -> str:
        raise NotImplementedError()

    def get_info(self) -> dict[str, Any]:
        raise NotImplementedError()

    def get_src_width(self) -> int:
        raise NotImplementedError()

    def get_src_height(self) -> int:
        raise NotImplementedError()

    def get_src_format(self) -> str:
        raise NotImplementedError()

    def get_dst_width(self) -> int:
        raise NotImplementedError()

    def get_dst_height(self) -> int:
        raise NotImplementedError()

    def get_dst_format(self) -> str:
        raise NotImplementedError()

    def convert_image(self, image: ImageWrapper) -> ImageWrapper:
        raise NotImplementedError()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.client.base.stub import StubClientMixin
from xpra.util.env import envbool
from xpra.log import Logger

log = Logger("client")

SPLASH_LOG = envbool("XPRA_SPLASH_LOG", False)


class ProgressClient(StubClientMixin):
    """
    Encapsulates functions for managing the splash screen
    """

    def __init__(self):
        self.progress_process = None

    def show_progress(self, pct: int, text="") -> None:
        pp = self.progress_process
        log(f"progress({pct}, {text!r}) progress process={pp}")
        if SPLASH_LOG:
            log.info(f"{pct:3} {text}")
        if pp:
            pp.progress(pct, text)

    def stop_progress_process(self, reason="closing") -> None:
        pp = self.progress_process
        if not pp:
            return
        self.show_progress(100, reason)
        self.progress_process = None
        if pp.poll() is not None:
            return
        from subprocess import TimeoutExpired
        try:
            if pp.wait(0.1) is not None:
                return
        except TimeoutExpired:
            pass
        try:
            pp.terminate()
        except OSError:
            pass

    def cleanup(self) -> None:
        self.stop_progress_process()



# This file is part of Xpra.
# Copyright (C) 2010 Antoine Martin <antoine@xpra.org>
# Xpra is released under the terms of the GNU GPL v2, or, at your option, any
# later version. See the file COPYING for details.

from xpra.server.subsystem.stub import StubServerMixin
from xpra.x11.error import xsync
from xpra.log import Logger

log = Logger("server", "tray")


class SystemTrayServer(StubServerMixin):

    def __init__(self):
        StubServerMixin.__init__(self)
        self.system_tray = False
        self._tray = None

    def init(self, opts) -> None:
        self.system_tray = opts.system_tray

    def setup(self) -> None:
        if self.system_tray:
            self.init_system_tray()

    def cleanup(self) -> None:
        tray = self._tray
        if tray:
            self._tray = None
            tray.cleanup()

    def init_system_tray(self) -> None:
        try:
            with xsync:
                from xpra.x11.tray import SystemTray
                self._tray = SystemTray()
        except RuntimeError as e:
            log("init_system_tray()", exc_info=True)
            log.error("Error: unable to setup system tray forwarding")
            log.estr(e)
